import pytest

from dli.client.builders import DatasetBuilder
from dli.client.exceptions import CatalogueEntityNotFoundException
from tests.integration import random_name


@pytest.mark.xfail(
    reason='The packages() and datasets() endpoints are disabled')
@pytest.mark.integration
def test_package_search(client):
    packages = client.packages(
        search_term='0_MS_2019_08_19', only_mine=False
    )

    assert packages

    packages = client.packages(
        search_term='0_MS_2019_08_19', only_mine=True
    )

    assert packages


@pytest.mark.xfail(
    reason='The packages() and datasets() endpoints are disabled')
@pytest.mark.integration
def test_dataset_search(client, csv_dataset_builder):

    with client.with_accounts([
        'bdf5fce5-3891-4ec2-9752-2dd89ae37e03',
        '11390b62-bb5e-4851-b129-7cd62442c8f8'
    ]) as other_client:
        package = other_client.register_package(
            name=random_name(),
            description="my package description",
            topic="Automotive",
            access="Restricted",
            internal_data="Yes",
            data_sensitivity="Public",
            terms_and_conditions="Terms",
            publisher="my publisher",
            access_manager_id='11390b62-bb5e-4851-b129-7cd62442c8f8',
            tech_data_ops_id='bdf5fce5-3891-4ec2-9752-2dd89ae37e03',
            manager_id='bdf5fce5-3891-4ec2-9752-2dd89ae37e03'
        )


        dataset = other_client.register_dataset(
            DatasetBuilder(
                package_id=package.package_id,
                name='dataset-files-test-' + random_name(),
                description="My dataset description",
                content_type="Pricing",
                data_format='CSV',
                publishing_frequency="Weekly",
                taxonomy=[]
            ).with_external_storage(
                location="jdbc://connectionstring:1232/my-db"
            )
        )

    assert not client.get_dataset(dataset.id).has_access
    assert not client.datasets(only_mine=True, search_term=dataset.id)
    assert client.datasets(only_mine=False, search_term=dataset.id)


@pytest.mark.integration
def test_package_get(client, package):
     assert client.packages._get(package.name)


@pytest.mark.integration
def test_dataset_get_returns_exception_on_bad_name(client):
    # TODO: extract Dataset tests to new file.
    with pytest.raises(CatalogueEntityNotFoundException):
        client.datasets._get('not a short code')


@pytest.mark.integration
def test_packages_get_returns_exception_on_bad_name(client):
    # TODO: extract Packages tests to new file.
    with pytest.raises(CatalogueEntityNotFoundException):
        client.packages._get('not a short code')



import pytest

from dli.client.exceptions import CatalogueEntityNotFoundException


@pytest.mark.integration
def test_files_method_in_dataset_with_qa_env(csv_dataset):
    for x in csv_dataset.instances.all():
        print('csv_dataset', x)
    c = csv_dataset.instances.latest()
    c_files = c.files()
    assert c_files is not None


@pytest.mark.integration
@pytest.mark.xfail
def test_dataset_partitions(csv_dataset):
    partitions = csv_dataset._partitions()
    assert set(partitions['as_of_date']) == set(
        '2020-01-27', '2020-01-28', '2020-01-29', '2020-01-30', '2020-01-31'
    )
    assert partitions['location'] == ['us']
    assert partitions['as_of_date'] == ['2000-01-01']


@pytest.mark.integration
def test_files_method_in_empty_dataset_with_qa_env(empty_dataset):
    with pytest.raises(CatalogueEntityNotFoundException):
        empty_dataset.instances.latest().files()


@pytest.mark.integration
def test_empty_package_with_qa_env(package, client):
    client.get_package(name=package.name)



import multiprocessing
import warnings
from concurrent.futures.thread import ThreadPoolExecutor
from urllib.parse import urljoin
import humps  # noqa: I900

from dli.models.paginator import Paginator
from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls, dataset_urls
from dli.models import log_public_functions_calls_using, AttributesDict

THREADS = multiprocessing.cpu_count()

@log_public_functions_calls_using(
    [analytics_decorator, logging_decorator], class_fields_to_log=['datafile_id']
)
class InstanceModel(AttributesDict):

    def __init__(self, **kwargs):
        # Ignore the datafile's files
        kwargs.pop('files', [])
        super().__init__(**kwargs)

    def files(self):
        """
        :return: list of file models for files in the instance.
        """
        url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_manifest.format(
                id=self.datafile_id
            )
        )

        response = self._client.session.get(url)
        return [
            self._client._File(
                datafile_id=self.datafile_id,
                **d['attributes']
            )
            for d in response.json()['data']
        ]

    @classmethod
    def _from_v1_entity(cls, entity):
        properties = humps.decamelize(entity['properties'])
        return cls(**properties)

    def download_all(self, to='./', flatten=False):
        """
        This method will soon become deprecated. Please use `download`
        function instead.

        :param str to: The path on the system, where the files
            should be saved. must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]


        :return: List of paths to the downloaded files.
        """
        warnings.warn(
            'This method will soon become deprecated. '
            'Please use `download` function instead.',
            PendingDeprecationWarning
        )
        return self.download(to, flatten)

    def download(self, to='./', flatten=False):
        """
        Download the files for the instance, then return a list of the file
        paths that were written.

        :param str to: The path on the system, where the files
            should be saved. must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]


        :return: List of paths to the downloaded files.
        """
        files = self.files()

        def _download(file_):
            return file_.download(to=to, flatten=flatten)

        threads = min(THREADS, max(1, len(files)))
        with ThreadPoolExecutor(max_workers=threads) as executor:
            return list(executor.map(_download, files))

    def __str__(self):
        separator = "-"*80
        return f"\nINSTANCE {self.datafile_id}"


@log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['_dataset.dataset_id']
)
class InstancesCollection(AttributesDict):

    def __init__(self, dataset=None):
        self._dataset = dataset
        self._paginator = Paginator(
            dataset_urls.datafiles.format(id=self._dataset.id),
            self._client.Instance, self._client.Instance._from_v1_entity
        )

    def latest(self):
        """:return: The latest instance."""
        response = self._client.session.get(
            dataset_urls.latest_datafile.format(id=self._dataset.id)
        ).json()

        return self._client.Instance._from_v1_entity(response)

    def all(self):
        """:return: All the instances."""
        warnings.warn(
            'The result of calling `.all` will be cached. If you want fresh '
            'results the next time you call `.all`, then please re-create the '
            'dataset variable before calling `.all`.'
        )
        yield from self._paginator



import os
import tempfile
import pytest

from moto import mock_s3
from datetime import datetime, timedelta
from unittest import TestCase
from unittest.mock import patch, Mock
from boto3 import Session
from botocore.exceptions import ClientError
from freezegun import freeze_time

from dli.client.exceptions import InsufficientPrivilegesException, S3FileDoesNotExist
from dli.client.s3 import Client, S3DatafileWrapper
from tests.localstack_helper import _fake_aws_credentials


session = Session(**_fake_aws_credentials())


@pytest.mark.integration
class S3ClientTestCase(TestCase):

    def setUp(self):
        self.target = Client(session)
        bucket = self.target.s3_resource.Bucket('bucket')
        bucket.create()

        def cleanup():
            bucket.objects.all().delete()
            bucket.delete()

        self.addCleanup(cleanup)

    def test_download_file_validates_destination_is_a_directory(self):
        self.target.s3fs.put(__file__, "s3://bucket/file")
        with self.assertRaises(NotADirectoryError):
            with tempfile.NamedTemporaryFile() as cm:
                self.target.download_files_from_s3_path("s3://bucket/file", cm.name)

    def test_download_file_creates_destination_directory_if_it_does_not_exist(self):
        self.target.s3fs.put(
            __file__,
            os.path.join("s3://bucket/location/", os.path.basename(__file__))
        )

        with tempfile.TemporaryDirectory() as dest:
            dest = os.path.join(dest, "dir1", "dir2")
            self.target.download_files_from_s3_path(
                "s3://bucket/location/test_s3_client.py",
                dest
            )

            self.assertTrue(os.path.exists(dest) and os.path.isdir(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "location", "test_s3_client.py")))

    def test_download_file_can_download_folders(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/file2.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/subdir/file3.txt")

        with tempfile.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "file2.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "subdir", "file3.txt")))

    def test_download_file_can_download_single_file(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")

        with tempfile.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf/file1.txt", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))


@pytest.mark.integration
class S3DatafileWrapperTestCase(TestCase):

    def setUp(self):
        self.client = Client(session)
        self.s3fs = self.client.s3fs
        bucket = self.client.s3_resource.Bucket('bucket')
        bucket.create()

        def cleanup():
            bucket.objects.all().delete()
            bucket.delete()

        self.addCleanup(cleanup)

    def test_files_returns_files_in_datafile(self):
        files = [
            "s3://bucket/tfrfid/a",
            "s3://bucket/tfrfid/b",
            "s3://bucket/tfrfid/c",
        ]

        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)
        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_files_will_recurse_directories(self):
        files = [
            "s3://bucket/tfwrd/a/aa",
            "s3://bucket/tfwrd/b/bb/bbb",
            "s3://bucket/tfwrd/c/cc/ccc/cccc1",
            "s3://bucket/tfwrd/c/cc/ccc/cccc2",
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [
                {'path': 's3://bucket/tfwrd/a'},
                {'path': 's3://bucket/tfwrd/b'},
                {'path': 's3://bucket/tfwrd/c'}
            ]
        }

        target = S3DatafileWrapper(datafile, self.client)

        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_only_files_in_datafile_are_displayed(self):
        files = [
            "s3://bucket/tofidad/a",
            "s3://bucket/tofidad/b",
            "s3://bucket/tofidad/c",
            "s3://bucket/tofidad/d"
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files[0:1]]
        }
        target = S3DatafileWrapper(datafile, self.client)
        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files[0:1]]
        )

    def test_can_open_file(self):
        files = [
            "s3://bucket/tcof/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)
            with self.s3fs.open(f, mode="wb") as s3f:
                s3f.write(b"test 1")
                s3f.flush()

        datafile = {
            "files":  [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)

        with target.open_file("bucket/tcof/a") as s3file:
            assert s3file is not None
            assert s3file.read() == b"test 1"

    def test_unknown_file_is_handled_gracefully(self):
        files = [
            "s3://bucket/tufihg/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)

        with self.assertRaises(S3FileDoesNotExist):
            res = target.open_file("bucket/unknown/file")

@pytest.mark.integration
class S3UploadTestCase(TestCase):

    def setUp(self):
        self.patcher = patch('s3fs.S3FileSystem')
        self.mock_s3fs_instance = self.patcher.start().return_value
        client = Client(session)
        self.s3_client = client

        self.bucket = client.s3_resource.Bucket('s3-test')
        self.bucket.create()

        def cleanup():
            self.bucket.objects.all().delete()
            self.bucket.delete()

        self.addCleanup(cleanup)

        self.s3_location = 's3-test/temp/'

    def test_upload_files_to_s3_normal(self):
        # create some dummy files
        temp_file_list = [tempfile.NamedTemporaryFile() for i in range(3)]
        temp_file_path_list = [f.name for f in temp_file_list]

        expected_s3_location_list = [{
            "path": "s3://{}{}".format(self.s3_location, os.path.basename(file)),
            "size": 0
        } for file in temp_file_path_list]

        upload_result = self.s3_client.upload_files_to_s3(temp_file_path_list, self.s3_location)
        for temp_file in temp_file_list:
            temp_file.close()

        assert len(list(self.bucket.objects.all())) == 3
        self.assertCountEqual(upload_result, expected_s3_location_list)

    def test_upload_files_with_no_bucket_access_fails(self):
        error = ClientError(
            {'Error': {'Code': 'AccessDenied', 'Message': 'Access Denied'}},
            'NOOP'
        )

        with patch.object(self.s3_client, 's3_client') as patched_s3_client:
            patched_s3_client.upload_file.side_effect = (
                error
            )

            with self.assertRaises(InsufficientPrivilegesException):
                self.s3_client.upload_files_to_s3([os.path.relpath(__file__)], self.s3_location)

    def tearDown(self):
        self.patcher.stop()


@mock_s3
def test_refresh_credentials():
    client = Mock()
    client.get_s3_access_keys_for_dataset.return_value = Mock(
        access_key_id='1234',
        secret_access_key='abc',
        session_token='xyz'
    )

    start = datetime(2010, 1, 1, 0, 0, 0)

    with freeze_time(start) as frozen_datetime:
        s3_client = Client.from_dataset(client, '1234')
        s3 = s3_client.boto_session.client('s3')

        assert client.get_s3_access_keys_for_dataset.call_count == 1

        s3.list_buckets()

        assert client.get_s3_access_keys_for_dataset.call_count == 1
        frozen_datetime.tick(delta=timedelta(hours=2))

        s3_client.boto_session.client('s3').list_buckets()
        assert client.get_s3_access_keys_for_dataset.call_count == 2



import logging
import uuid

import pytest

from tests.common import SdkIntegrationTestCase, eventually

logger = logging.getLogger(__name__)


@pytest.mark.integration
class SearchTestCase(SdkIntegrationTestCase):

    def entities_are_returned_in_search(self, search_func, expected_min_count):
        def search():
            result = search_func()
            self.assertTrue(len(result) >= expected_min_count)
            return result

        return eventually(search)

    def test_search_all_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search(term='searchable', count=c))

    def test_search_packages_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search_packages(term='searchable', count=c))

    def test_search_special_chars(self):
        payload = {
            "name": 'searchable package' + "_" + str(uuid.uuid4()),
            # hello chinese world
            "description":  "你好，世界",
            "publisher":  "datalake-mgmt",
            "manager_id":  "datalake-mgmt",
            "access_manager_id":  "datalake-mgmt",
            "tech_data_ops_id":  "datalake-mgmt",
            "topic":  "Academic/Education",
            "access":  'Restricted',
            "internal_data":  "No",
            "data_sensitivity":  "Public",
            "terms_and_conditions":  "Terms and Conditions"
        }
        self.client.register_package(**payload).package_id
        assert self.entities_are_returned_in_search(
            lambda: self.client.search_packages('你好'), 1
        )

    def test_search_packages(self):
        self.assertEqual(len(self.client.search_packages(term="")), 0)

        self.create_package("searchable package")
        search_term = "searchable"

        results = self.entities_are_returned_in_search(lambda: self.client.search_packages(search_term), 1)
        self.assertTrue(all([hasattr(package, "package_id") and search_term in package.name for package in results]))

    def test_search_datasets_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search_datasets(term='searchable', count=c))

    def test_search_datasets(self):
        self.assertEqual(len(self.client.search_datasets(term="")), 0)

        package_id = self.create_package("package")
        builder = self.dataset_builder(package_id, "searchable dataset")
        builder = builder.with_external_storage(location="jdbc://somewhere")
        self.client.register_dataset(builder)

        search_term = "searchable"
        results = self.entities_are_returned_in_search(lambda: self.client.search_datasets(search_term), 1)
        for result in results:
            self.assertEqual(result.type, "dataset")
            self.assertIn(search_term, result.name)



from requests_toolbelt.adapters import host_header_ssl

from dli import __version__


class DLIAdapter(host_header_ssl.HostHeaderSSLAdapter):

    def __init__(self, session, *args, **kwargs):
        self.session = session
        super().__init__(*args, **kwargs)

    def add_headers(self, request, **kwargs):
        request.headers["X-Data-Lake-SDK-Version"] = str(__version__)
        # if a host has been provided, then we need to set it on the header
        if self.session.host:
            request.headers['Host'] = self.session.host

        super().add_headers(request, **kwargs)


class DLIBearerAuthAdapter(DLIAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)
        if self.session.auth_key and 'Authorization' not in request.headers:
            request.headers['Authorization'] = f'Bearer {self.session.auth_key}'


class DLISirenAdapter(DLIAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)
        request.headers['Content-Type'] = "application/vnd.siren+json"


class DLICookieAuthAdapter(DLIAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)
        # Accounts V1 authentication is broken, in that it only accepts
        # a cookie rather than an API key.
        request.headers['Cookie'] = f'oidc_id_token={self.session.auth_key}'


class DLIAccountsV1Adapter(DLISirenAdapter, DLICookieAuthAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)


class DLIInterfaceV1Adapter(DLISirenAdapter, DLIBearerAuthAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)


class DLISamAdapter(DLIAdapter):
    def add_headers(self, request, **kwargs):
        super().add_headers(request, **kwargs)
        request.headers['Content-Type'] = "application/x-www-form-urlencoded"




import os

import pytest
import tempfile

from unittest.mock import MagicMock, call, patch
from dli.client.s3 import Client
from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    DatalakeException,
    DownloadFailed,
)
from tests.common import SdkIntegrationTestCase


@pytest.mark.integration
class DatafileTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super().setUp()

        self.package_id = self.create_package("test_datafile_functions")
        self.dataset_name = "test_datafile_functions"
        self.builder = self.dataset_builder(
            self.package_id, self.dataset_name)

    def create_s3_dataset(self, bucket_name, prefix="prefix"):
        return self.register_s3_dataset(self.package_id, self.dataset_name, bucket_name, prefix=prefix)

    def create_dummy_datafile(self):
        dataset = self.register_s3_dataset(self.package_id, self.dataset_name, "test-bucket", "prefix")
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "dummy",
            [os.path.relpath(__file__)],
            "prefix",
            data_as_of='2000-01-01'
        )

        return datafile

    def test_get_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_datafile("unknown")

    def test_get_unknown_s3_datafile_returns_error(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_s3_datafile("unknown")

    def test_can_get_s3_datafile(self):
        s3_dataset = self.create_s3_dataset("test-bucket")
        datafile = self.client.register_s3_datafile(
            s3_dataset.dataset_id,
            "test_get_s3_datafile",
            [os.path.relpath(__file__)],
            "prefix",
            data_as_of='2000-01-01'
        )


        s3_datafile = self.client.get_s3_datafile(datafile.datafile_id)
        self.assertEqual(s3_datafile.datafile_id, datafile.datafile_id)

    def test_register_datafile_metadata(self):
        files = [{'path': "/path/to/file/A", 'size': 99999},
                 {'path': "/path/to/file/B", 'size': 88888}]
        dataset = self.client.register_dataset(
            self.builder.with_external_storage(
                location="jdbc://connectionstring:1232/my-db"
            )
        )
        datafile = self.client.register_datafile_metadata(
            dataset.dataset_id,
            "test_register_dataset_metadata",
            files,
            data_as_of='2000-01-01'
        )

        self.assertEqual(datafile.dataset_id, dataset.dataset_id)
        self.assertEqual(datafile.files, files)

    def test_register_datafile_metadata_fails_when_data_as_of_is_invalid(self):
        files = [{'path': "/path/to/file/A", 'size': 99999},
                 {'path': "/path/to/file/B", 'size': 88888}]
        dataset = self.client.register_dataset(
            self.builder.with_external_storage(
                location="jdbc://connectionstring:1232/my-db"
            )
        )

        with self.assertRaises(ValueError):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "test_register_dataset_metadata",
                files,
                data_as_of='2000-111-01'
            )

    def test_register_datafile_metadata_fails_when_no_files_provided(self):
        dataset = self.client.register_dataset(self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"))
        with self.assertRaises(Exception):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "test_register_datafile_metadata_fails_when_no_files_provided",
                files=[]
            )

    def test_update_datafile_fails_for_unknown_datafile(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_datafile_metadata("unknown")

    def test_register_s3_datafile_can_create_datafile_uploading_files(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
            file = os.path.relpath(__file__)  # upload ourselves as a dataset
            datafile = self.client.register_s3_datafile(
                dataset.dataset_id,
                "test_register_s3_datafile_can_create_datafile_uploading_files",
                [file],
                "prefix",
                data_as_of='2000-01-01'
            )


            self.assertIsNotNone(datafile)
            self.assertEqual(datafile.dataset_id, dataset.dataset_id)
            self.assertEqual(datafile.files, [
                {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                    os.path.basename(file)}
            ])

    def test_register_datafile_metadata_on_datalake_dataset(self):
        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        file = os.path.relpath(__file__)  # upload ourselves as a dataset
        datafile = self.client.register_datafile_metadata(
            dataset.dataset_id,
            "test_register_datafile_metadata_on_datalake_dataset",
            [
                {"path": "s3://dev-ihsm-dl-pkg-test/prefix/test"}
            ],
            data_as_of='2000-01-01'
        )

        self.assertIsNotNone(datafile)
        self.assertEqual(datafile.dataset_id, dataset.dataset_id)

    def test_update_datafile_metadata_fails_when_data_as_of_is_invalid(self):
        datafile = self.create_dummy_datafile()
        with self.assertRaises(ValueError):
            self.client.edit_datafile_metadata(
                datafile.datafile_id,
                name="correct name",
                data_as_of="dataAsOf"
            )

    def test_update_datafile_merges_changes_with_existing_datafile(self):
        datafile = self.create_dummy_datafile()
        updated = self.client.edit_datafile_metadata(
            datafile.datafile_id,
            name="correct name"
        )

        self.assertEqual(datafile.datafile_id, updated.datafile_id)
        self.assertEqual(datafile.dataset_id, updated.dataset_id)
        self.assertEqual(datafile.files, updated.files)
        self.assertEqual(updated.name, "correct name")

    def test_can_delete_datafile(self):
        datafile = self.create_dummy_datafile()
        # delete the datafile
        self.client.delete_datafile(datafile.datafile_id)
        # can't read it back
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_datafile(datafile.datafile_id)

    def test_delete_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_datafile("unknown")

    def test_can_add_files_to_existing_datafile(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            file = os.path.relpath(__file__)  # upload ourselves as a dataset
            file2 = '../test_sandbox/samples/data/AAPL.csv'

            dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
            datafile = self.client.register_s3_datafile(
                dataset.dataset_id,
                "test_can_add_files_to_existing_datafile",
                [file],
                "prefix",
                data_as_of='2000-01-01'
            )


            updated = self.client.add_files_to_datafile(
                datafile.datafile_id, 'prefix', [file2])

            # countEqual asserts that the collections match disrespect of order
            # talk about naming stuff
            self.assertCountEqual(
                [
                    {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                        os.path.basename(file)},
                    {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                        os.path.basename(file2)}
                ],
                updated.files
            )

    def test_add_files_to_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.add_files_to_datafile(
                'unknown', 'prefix', ["/path/to/file/A"])

    def test_register_datafile_to_nonexisting_dataset_raises_exception(self):
        with self.assertRaises(DatalakeException) as context:
            self.client.register_datafile_metadata(
                "unkknown",
                "test_register_datafile_to_nonexisting_dataset_raises_exception",
                [{'path': "/path/to/file/A"}, {'path': "/path/to/file/B"}],
                data_as_of='2000-01-01'
            )
            self.assertTrue(
                'Dataset with id unknown not found' in context.exception)


@pytest.mark.integration
class DownloadDatafileTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super().setUp()
        # create a package
        self.package_id = self.create_package("test download datafile")
        self.dataset_name = "test download datafile"
        self.builder = self.dataset_builder(
            self.package_id, self.dataset_name)

    def create_s3_dataset(self, bucket_name, prefix="prefix"):
        return self.register_s3_dataset(self.package_id, self.dataset_name, bucket_name, prefix)

    def test_download_datafile_for_unknown_datafile_fails(self):
        with self.assertRaises(Exception):
            self.client.download_datafile("unknown")

    def test_download_datafile_retrieves_all_files_in_datafile(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
            datafile = self.client.register_s3_datafile(
                dataset.dataset_id,
                "test_download_dataset_retrieves_all_files_in_dataset",
                [
                    '../test_sandbox/samples/data/AAPL.csv',
                    '../test_sandbox/samples/data/MSFT.csv'
                ],
                "prefix",
                '2000-01-01'
            )

            with patch('dli.client.s3.Client.download_files_from_s3_path') \
                    as s3_download:
                with tempfile.TemporaryDirectory() as dest:
                    self.client.download_datafile(
                        datafile.datafile_id,
                        dest,
                        flatten=False
                    )

                    # validate we got the expected calls
                    s3_download.assert_has_calls([
                        call(
                            destination=dest,
                            flatten=False,
                            s3_path="s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv",
                        ),
                        call(
                            destination=dest,
                            flatten=False,
                            s3_path = "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
                        ),
                    ], any_order=True)

    def test_download_dataset_keeps_going_if_a_file_in_the_dataset_fails(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            def _download(_, file, dest):
                if file.endswith("AAPL.csv"):
                    raise Exception("")

            dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
            datafile = self.client.register_s3_datafile(
                dataset.dataset_id,
                "test_download_dataset_keeps_going_if_a_file_in_the_dataset_fails",
                [
                    '../test_sandbox/samples/data/AAPL.csv',
                    '../test_sandbox/samples/data/MSFT.csv'
                ],
                "prefix",
                '2000-01-01'
            )


            with self.assertRaises(DownloadFailed):
                with patch('dli.client.s3.Client.download_files_from_s3_path', _download) as s3_download:
                    with tempfile.TemporaryDirectory() as dest:
                        self.client.download_datafile(datafile.datafile_id, dest)

                        # validate we got the expected calls
                        s3_download.assert_has_calls([
                            call("s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv", dest),
                            call("s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv", dest),
                        ],
                            any_order=True
                        )


@pytest.mark.integration
class RegisterDatafileTestCase(SdkIntegrationTestCase):

    def set_s3_client_mock(self):
        """This prevents the s3 client from being mocked"""

    def create_and_get_dataset_id(self):
        package_id = self.create_package(
            "RegisterDatafileTestCase"
        )
        dataset = self.register_s3_dataset(
            package_id, "test", "dev-ihsm-dl-pkg-test"
        )

        return dataset.dataset_id

    def test_can_upload_datafile_when_provided_folder_with_relative_path(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_when_provided_folder_with_relative_path",
            [sample_data],
            "prefix",
            '2000-01-01'
        )


        # assert the files were uploaded and that
        # their sizes have been resolved
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(os.path.join(sample_data, "AAPL.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv",
            "size": os.path.getsize(os.path.join(sample_data, "MSFT.csv"))
        },
            datafile.files
        )

    def test_can_upload_datafile_when_provided_s3_prefix_without_trailing_slash(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        # Prefix Without trailing slash
        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_when_provided_s3_prefix_with_or_without_trailing_slashes 1",
            [sample_data],
            "prefix",
            '2000-01-01'
        )


        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(sample_data)
        },
            datafile.files
        )

    def test_can_upload_datafile_with_data_as_of_date_not_in_iso_format(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_with_data_as_of_date_not_in_iso_format 1",
            [sample_data],
            "prefix",
            '3210-1-1'
        )


        assert self.client.get_datafile(datafile.datafile_id).data_as_of == '3210-01-01'
        assert {
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(sample_data)
        } in datafile.files

    def test_cannot_upload_datafile_with_invalid_data_as_of(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        with self.assertRaises(ValueError):
            self.client.register_s3_datafile(
                self.create_and_get_dataset_id(),
                "test_can_upload_datafile_with_data_as_of_date_not_in_iso_format 1",
                [sample_data],
                "prefix",
                '2019-11-123'
            )

    def test_can_upload_datafile_files_recursively(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/stocks'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_files_recursively",
            [sample_data],
            "prefix",
            '2000-01-01'
        )


        # assert the files were uploaded and that
        # their sizes have been resolved
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/readme.txt",
            "size": os.path.getsize(os.path.join(sample_data, "readme.txt"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/MSFT.csv",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/MSFT.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/AAPL.csv",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/AAPL.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/cortana/master.txt",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/cortana/master.txt"))
        },
            datafile.files
        )


    def test_download_datafile_should_keep_folder_structure_as_in_datafile(self):
        download_file = MagicMock()

        class TestS3Client(Client):
            def __init__(self, *args):
                super().__init__(*args)
                self.s3_client.download_file = download_file

            def _list_objects(self, bucket, prefix):
                return [{
                    'Key': prefix
                }]

        with patch('dli.client.components.datafile.Client', TestS3Client) as client, \
             patch('dli.client.s3.os.makedirs') as makedirs_mock:
            sample_data = os.path.join(
                os.path.dirname(__file__),
                'resources/stocks'
            )

            datafile = self.client.register_s3_datafile(
                self.create_and_get_dataset_id(),
                "test_download_datafile_should_keep_folder_structure_as_in_datafile",
                [sample_data],
                "prefix",
                '2000-01-01'
            )


            self.client.download_datafile(datafile.datafile_id, '/var/abc/')

            directories_created = [
                d[0][0] for d in makedirs_mock.call_args_list
            ]

            files_created = [
                f[0][2] for f in download_file.call_args_list
            ]

            assert directories_created == [
                '/var/abc/prefix',
                '/var/abc/prefix/microsoft',
                '/var/abc/prefix/microsoft',
                '/var/abc/prefix/microsoft/cortana'
            ]

            assert files_created == [
                '/var/abc/prefix/readme.txt',
                '/var/abc/prefix/microsoft/MSFT.csv',
                '/var/abc/prefix/microsoft/AAPL.csv',
                '/var/abc/prefix/microsoft/cortana/master.txt'
            ]

    def test_download_datafile_returns_paths_where_files_were_downloaded(self):
        download_file = MagicMock()

        class TestS3Client(Client):
            def __init__(self, *args):
                super().__init__(*args)
                self.s3_client.download_file = download_file

            def _list_objects(self, bucket, prefix):
                return [{
                    'Key': prefix
                }]

        with patch('dli.client.components.datafile.Client',
                   TestS3Client) as client, \
                patch('dli.client.s3.os.makedirs') as makedirs_mock:
            sample_data = os.path.join(
                os.path.dirname(__file__),
                'resources/stocks'
            )

            datafile = self.client.register_s3_datafile(
                self.create_and_get_dataset_id(),
                "test_download_datafile_should_keep_folder_structure_as_in_datafile",
                [sample_data],
                "prefix",
                '2000-01-01'
            )


            returned_list_of_files = self.client.download_datafile(
                datafile.datafile_id,'/var/abc/'
            )

            files_created = [
                '/var/abc/prefix/readme.txt',
                '/var/abc/prefix/microsoft/MSFT.csv',
                '/var/abc/prefix/microsoft/AAPL.csv',
                '/var/abc/prefix/microsoft/cortana/master.txt'
            ]
            assert returned_list_of_files == files_created

    def test_flattened_download_datafile_returns_one_path(self):
        download_file = MagicMock()

        class TestS3Client(Client):
            def __init__(self, *args):
                super().__init__(*args)
                self.s3_client.download_file = download_file

            def _list_objects(self, bucket, prefix):
                return [{
                    'Key': prefix
                }]

        with patch('dli.client.components.datafile.Client',
                   TestS3Client) as client, \
                patch('dli.client.s3.os.makedirs') as makedirs_mock:
            sample_data = os.path.join(
                os.path.dirname(__file__),
                'resources/stocks'
            )

            datafile = self.client.register_s3_datafile(
                self.create_and_get_dataset_id(),
                "test_download_datafile_should_flatten_folder_structure",
                [sample_data],
                "prefix",
                '2000-01-01'
            )


            returned_list_of_files = self.client.download_datafile(
                datafile.datafile_id, '/var/abc/', flatten=True)

            files_created = [ # package_dataset_asof_filename
                '/var/abc/readme.txt',
                '/var/abc/MSFT.csv',
                '/var/abc/AAPL.csv',
                '/var/abc/master.txt'
            ]
            assert sorted(returned_list_of_files) == sorted(files_created)



import json
import os
import time
from urllib.parse import urlencode

import keyring
from keyring.errors import NoKeyringError
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait
import pytest
import unittest
import jwt
from unittest.mock import MagicMock, patch
import httpretty

from dli import __version__, __product__
from dli.client.dli_client import Session, DliClient
from dli.client.environments import _Environment
from dli.client.listener import _Listener
from dli.client.session import start_session
from dli.client import session
from dli.client.exceptions import (
    DatalakeException,
    InsufficientPrivilegesException,
    UnAuthorisedAccessException
)
from dli.siren import PatchedSirenBuilder

environ = MagicMock(catalogue='http://catalogue.local',
                    accounts='')
valid_token = jwt.encode({"exp": 9999999999}, 'secret')
expired_token = jwt.encode({"exp": 1111111111}, 'secret')


class SessionTestCase(unittest.TestCase):

    def test_can_decode_valid_jwt_token(self):
        ctx = Session(
            None,
            "key",
            environ,
            None,
            valid_token
        )

        self.assertFalse(ctx.has_expired)

    def test_can_detect_token_is_expired(self):
        ctx = Session(
            None,
            "key",
            environ,
            None,
            expired_token
        )
        self.assertTrue(ctx.has_expired)


class SessionRequestFactoryTestCase(unittest.TestCase):

    @pytest.fixture(autouse=True)
    def session(self):
        self.session = Session(None, None, environ, None, valid_token)

    @httpretty.activate
    def test_response_403_raises_InsufficientPrivilegesException(self):
        response_text = 'Insufficient Privileges'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=403, body=response_text
        )

        with self.assertRaises(InsufficientPrivilegesException):
            self.session.get('/test')

    @httpretty.activate
    def test_response_401_raises_UnAuthorisedAccessException(self):
        response_text = 'UnAuthorised Access'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=401, body=response_text
        )


        with self.assertRaises(UnAuthorisedAccessException):
            self.session.get('/test')

    @httpretty.activate
    def test_response_500_raises_DatalakeException(self):
        response_text = 'Datalake server error'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=500, body=response_text
        )

        with self.assertRaises(DatalakeException):
            self.session.get('/test')

    @httpretty.activate
    def test_sdk_version_is_included_in_header(self):
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/__api/',
            status=200, body="response"
        )
        # issue a request
        self.session.get('/__api/')

        request = httpretty.last_request()
        self.assertTrue("X-Data-Lake-SDK-Version" in request.headers)
        self.assertEqual(request.headers["X-Data-Lake-SDK-Version"], str(__version__))


class TestEnv:

    def __init__(self):
        self.sam = "http://sam.local"
        self.catalogue = "http://catalogue.local"
        self.accounts = "http://catalogue.local"
        self.consumption = "http://consumption.local"
        self.sam_client = "CLIENT_ID"


class TestSessionMock(Session):

    def __init__(self, id, pasw):
        self._get_SAM_auth_key = MagicMock()
        self._get_auth_key = MagicMock()
        self._get_decoded_token = MagicMock()
        self._get_expiration_date = MagicMock()
        self._set_mount_adapters = MagicMock()
        super().__init__(id, pasw, TestEnv(), "Test")


class TestClientMock(DliClient):

    def __init__(self, api_root, host=None, debug=None, strict=True,
                 access_id=None, secret_key=None, use_keyring=True):
        super().__init__("Test", access_id=access_id, secret_key=secret_key, use_keyring=use_keyring)

    def _new_session(self):
        return TestSessionMock("Test", "Test")


class TestSessionMockWithAuth(Session):

    def __init__(self, id, pasw):
        self._set_mount_adapters = MagicMock()
        self.access_id = id
        self.secret_key = pasw
        self.auth_key = None
        self._environment = TestEnv()
        self.host = None
        self.use_keyring = True
        self.siren_builder = PatchedSirenBuilder()


class TestClientMockWithAuth(TestClientMock):

    def _new_session(self):
        return TestSessionMockWithAuth(self.access_id, self.secret_key)


class TestSessionMockNoPrompt(Session):

    # required to prevent the web popup that we want to
    # launch ourselves / use selenium
    def __init__(self, id, pasw, env):
        super().__init__(id, pasw, env, "Test", auth_prompt=False)


class TestClientMockNoPrompt(DliClient):

    def _new_session(self):
        return TestSessionMockNoPrompt(None, None, self._environment)


@pytest.fixture
def clear_keyring(monkeypatch):
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: None)
    monkeypatch.setattr(keyring, 'set_password', lambda _, app, val: None)

@pytest.fixture
def mock_session(clear_keyring):
    yield TestSessionMock

@pytest.fixture
def mock_del_env_user(monkeypatch):
    monkeypatch.delenv("DLI_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("DLI_SECRET_ACCESS_KEY", raising=False)

@pytest.fixture
def real_client(request):
    # This is the S3 address used for the QA environment.
    api_root = os.environ[f'{request.param[0]}_API_URL']
    accessid = os.environ[f'{request.param[0]}_ACCESS_ID']
    secretkey = os.environ[f'{request.param[0]}_SECRET_ACCESS_KEY']
    client = None

    if request.param[1] == "Credentials":
        client = DliClient(
            api_root=api_root,
            access_id=accessid,
            secret_key=secretkey
        )
    elif request.param == 'API':
        pass
    else:
        pass

    yield client


@pytest.fixture
def mock_create_env_user(monkeypatch):
    monkeypatch.setenv("DLI_ACCESS_KEY_ID", "TestingUser")
    monkeypatch.setenv("DLI_SECRET_ACCESS_KEY", "TestingPass")


def test_credentials_no_api_key(monkeypatch, mock_create_env_user, recwarn):
    # test flow when u/p set and no api key set
    monkeypatch.setattr(DliClient, '_new_session', lambda x: x)
    dl = start_session()
    assert(issubclass(type(dl), DliClient))
    assert(dl.secret_key is not None and dl.access_id is not None)


def test_credentials_session(mock_create_env_user, mock_session):
    # test flow when u/p set and no api key set

    sesh = mock_session(os.environ["DLI_ACCESS_KEY_ID"],
                        os.environ["DLI_SECRET_ACCESS_KEY"])
    assert sesh._get_SAM_auth_key.call_count == 1


def test_api_key_deprecation_warning(recwarn, monkeypatch):
    # test flow when u/p set and api key set
    # api key should override
    monkeypatch.setattr(DliClient, '_new_session', MagicMock())
    dl = start_session(api_key="Test")
    assert(issubclass(type(dl), DliClient))
    assert "`api_key` will be deprecated in the future" \
           in ",".join([str(x.message) for x in recwarn.list])
    assert(dl.secret_key is not None and dl.access_id is None)


def test_credentials_and_api_key(mock_create_env_user, monkeypatch):
    # test flow when u/p set and api key set
    # api key should override
    monkeypatch.setattr(DliClient, '_new_session',  MagicMock())
    dl = start_session(api_key="Test")
    assert (dl.secret_key is not None and dl.access_id is None)


def test_api_session(mock_create_env_user, mock_session):
    # test flow when u/p set and no api key set
    # (there will be no acces id set)
    sesh = mock_session(None, "Test")
    assert sesh._get_auth_key.call_count == 1


def test_credentials_and_no_api_key(mock_create_env_user, monkeypatch, caplog):
    # test flow when no credentials and api key set
    monkeypatch.setattr(session, "get_client", lambda: TestClientMock)
    dl = start_session()
    for x in caplog.records:
        assert("old" not in x.message)


def test_no_credentials_and_api_key(mock_del_env_user, monkeypatch, caplog):
    # test flow when no credentials and api key set
    monkeypatch.setattr(session, "get_client", lambda: TestClientMock)
    dl = start_session(api_key="Test")
    assert (dl.secret_key is not None and dl.access_id is None)
    for x in caplog.records:
        assert("new" not in x.message)


def test_api_key_auth(mock_del_env_user, monkeypatch):
    # test old auth process (_get_auth_key())
    api_response = valid_token
    monkeypatch.setattr(session, "get_client", lambda: TestClientMockWithAuth)
    dl = start_session(api_key="API")
    dl._session.post = MagicMock()
    dl._session.post.return_value.text = api_response

    dl._session._auth_init()
    assert(dl and dl._session.auth_key is not None)


def test_credentials_auth(mock_create_env_user, monkeypatch):
    # test new auth process (_get_SAM_auth_key())
    sam_response = {
        "access_token": valid_token,
        "token_type": "Bearer", "expires_in": 3599}
    catalogue_response = {
        "access_token": valid_token,
        "token_type": "Bearer", "expires_in": 3599}

    monkeypatch.setattr(session, "get_client", lambda: TestClientMockWithAuth)
    dl = start_session()
    dl._session.post = MagicMock()
    dl._session.post.return_value.json.side_effect = [
        sam_response, catalogue_response
    ]

    dl._session._auth_init()
    assert(dl and dl._session.auth_key is not None)



def real_drive(port, postbox, env):
    chrome_options = Options()
    # chrome_options.add_argument("--headless")

    user = "dl_manual_test_2@ihsmarkit.com"
    pw = "-0o_p}i-Uz%*PQy_"

    driver = webdriver.Chrome(executable_path='./chromedriver',
                              options=chrome_options)
    driver.get(f"{_Listener.LOCALHOST}:{port}/login?postbox={postbox}&{urlencode(env.__dict__)}")

    WebDriverWait(driver, 120).until(
        ec.visibility_of_element_located((By.ID, "emailAddress"))
    ).send_keys(user)

    driver.find_element(By.ID, "continueButton").click()

    WebDriverWait(driver, 10).until(
        ec.visibility_of_element_located((By.ID, "i0116"))
    ).send_keys(user)

    driver.find_element(By.ID, "idSIButton9").click()

    WebDriverWait(driver, 10).until(
        ec.visibility_of_element_located((By.ID, "passwordInput"))
    ).send_keys(pw)

    driver.find_element(By.ID, "submitButton").click()

    driver.find_element(By.ID, "idSIButton9").click()
    # need to add some time, else the client quits before we
    # complete localhost activity.
    time.sleep(2)
    driver.quit()


def real_drive_off_cliff(port):
    #takes down the Listener

    chrome_options = Options()
    chrome_options.add_argument("--headless")
    user = "dl_manual_test_2@ihsmarkit.com"
    pw = "-0o_p}i-Uz%*PQy_"
    driver = webdriver.Chrome(executable_path='./chromedriver',
                              chrome_options=chrome_options)
    driver.get(f"{_Listener.LOCALHOST}:{port}/shutdown")
    driver.quit()


def real_login_attempt(sesh):
    postbox = _Listener.run(port=8080)
    real_drive(8080, postbox, sesh._environment)
    time.sleep(5)
    assert _Listener.VALUES.get(postbox) is not None
    return postbox


def test_reload_JWT_keyring(monkeypatch):
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: valid_token.decode())

    dl = TestClientMockWithAuth(api_root="Test")
    dl._session._get_web_auth_key = MagicMock()
    dl._session._auth_init()
    assert dl._session._get_web_auth_key.call_count == 0


def test_no_reload_delete_expired_JWT_keyring(monkeypatch):

    dl = TestClientMockWithAuth(api_root="Test")
    # we definitely set it in the first place
    keyring.set_password(
        __product__, dl._session._environment.catalogue, valid_token.decode())
    # later 'expires'
    orig = keyring.get_password
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: expired_token.decode())

    dl._session._get_web_auth_key = MagicMock()
    dl._session._get_web_auth_key.return_value = valid_token
    dl._session._auth_init()
    assert dl._session._get_web_auth_key.call_count == 1
    assert(orig(__product__, dl._session._environment.catalogue) is None)


@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
def test_set_JWT_keyring_web_auth_key(monkeypatch):
    orig = keyring.get_password
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: None)
    dl = TestClientMockWithAuth(api_root="Test")

    def catch(port, postbox):
        _Listener.VALUES[postbox] = valid_token.decode("utf-8")

    dl._session._get_web_auth_key(callback=catch)
    assert(orig(__product__, dl._session._environment.catalogue)
           == valid_token.decode())


@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
def test_web_session(mock_del_env_user):

    sesh = Session(
        "test", "test", _Environment("https://catalogue-dev.udpmarkit.net"),
        "Test", auth_prompt=False, auth_key=None)

    real_login_attempt(sesh)
    time.sleep(1)

    assert(len(_Listener.VALUES) == len(set(_Listener.VALUES.VALUES())))


@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
def test_web_flow(clear_keyring, mock_del_env_user, monkeypatch):

    monkeypatch.setattr(session, "get_client", lambda: TestClientMockNoPrompt)
    dl = start_session(root_url="https://catalogue-dev.udpmarkit.net")

    # this is what get_web_auth_key is otherwise doing,
    # but we need to unblock to allow selenium to operate.
    # so we register a custom default callback to call selenium
    # rather than the web browser
    def cb(port, postbox):
        real_drive(port, postbox, dl._environment)

    with patch.object(Session._get_web_auth_key, '__defaults__', (cb,)):
        dl._session._auth_init()

    assert dl._session.auth_key is not None

@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
@pytest.mark.parametrize('real_client', [('QA', "Credentials")],
                         indirect=['real_client'])
def test_credentials_on_catalogue(clear_keyring, real_client):
    # test that new auth actually works and returns JWT

    assert(real_client.datasets._get(
        "autotestdatasetItsdeclarationRomanZimmermannandpeople"
    ) is not None)

@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
@pytest.mark.parametrize('real_client', [('QA', "Credentials")],
                         indirect=['real_client'])
def test_webflow_on_catalogue(monkeypatch, real_client):

    monkeypatch.setattr(session, "get_client", lambda: TestClientMockNoPrompt)
    dl = start_session(root_url="https://catalogue-dev.udpmarkit.net")

    # this is what get_web_auth_key is otherwise doing,
    # but we need to unblock to allow selenium to operate.
    # so we register a custom default callback to call selenium
    # rather than the web browser
    def cb(port, postbox):
        real_drive(port, postbox, dl._environment)

    with patch.object(Session._get_web_auth_key, '__defaults__', (cb,)):
        dl._session._auth_init()

    # todo - doesnt work with these u/p sam logins
    assert dl.datasets._get("Sheep") is not None

def test_no_system_keyring_installed(monkeypatch):
    keyring.core.set_keyring(keyring.core.fail.Keyring())
    monkeypatch.setattr("dli.client.dli_client.Session._reload_or_web_flow.__defaults__", (True,))
    with pytest.raises(NoKeyringError):
        dl = start_session(root_url="https://catalogue-dev.udpmarkit.net")

def test_no_system_keyring_ack(monkeypatch, capsys):
    def t(pb):
        _Listener.VALUES[pb] =  valid_token.decode("utf-8")
    keyring.core.set_keyring(keyring.core.fail.Keyring())
    monkeypatch.setattr("dli.client.dli_client.Session._get_web_auth_key.__defaults__", ( (lambda port, pb: t(pb)), ))
    dl = start_session(root_url="https://catalogue-dev.udpmarkit.net")
    captured = capsys.readouterr()
    assert "no keyring manager available" in captured.out


from unittest.mock import Mock

import pytest

import dli
from dli.analytics import AnalyticsHandler


@pytest.fixture
def dli_client():
    session = Mock()
    session.has_expired = False

    environment = Mock(consumption='http://someurl')
    return Mock(_session=session, _environment=environment)


class TestAnalyticsHandler:

    def test_properly_creates_handler(self):
        logger = Mock()
        dli_client = Mock(logger=logger)
        analytics_handler = AnalyticsHandler(dli_client)
        assert analytics_handler._app_name == 'SDK'
        assert analytics_handler._app_version == dli.__version__
        assert analytics_handler._client == dli_client
        assert analytics_handler._logger == logger

    def test_handles_internal_exceptions(self, dli_client):
        exception_to_be_thrown = ValueError()
        dli_client.session.post.side_effect = exception_to_be_thrown

        analytics_handler = AnalyticsHandler(dli_client)
        analytics_handler.create_event('user', 'org', properties={})
        dli_client.session.post.assert_called_once()
        analytics_handler._logger.exception.assert_called_once_with(
            'Error while sending analytics: ', exception_to_be_thrown
        )

    def test_sends_proper_message_on_proper_input(self, dli_client):
        analytics_handler = AnalyticsHandler(dli_client)
        analytics_handler.create_event('user', 'org', properties={})
        dli_client.session.post.assert_called_once_with(
            'http://someurl/analytics/', json={
                'data': [{'attributes': {
                    'application_name': 'SDK',
                    'application_version': dli.__version__,
                    'user_id': 'user',
                    'entity': None,
                    'action': None,
                    'organisation_id': 'org',
                    'result': None,
                    'properties': {}}}]
            }
        )

    def test_sends_message_with_filtered_out_properties(self, dli_client):
        analytics_handler = AnalyticsHandler(dli_client)
        properties = {
            'package_id': '1',
            'dataset_id': '2',
            'name': 'some_name',
            'dataset_name': 'd_name',
            'datafile_id': 'datafile_id',
            'dictionary_id': 'dictionary_id',
            'api_key': 'key',
            'unknown': 'property'
        }
        analytics_handler.create_event('user', 'org', properties=properties)
        properties.pop('unknown')

        dli_client.session.post.assert_called_once_with(
            'http://someurl/analytics/', json={
                'data': [{'attributes': {
                    'application_name': 'SDK',
                    'application_version': dli.__version__,
                    'user_id': 'user',
                    'entity': None,
                    'action': None,
                    'organisation_id': 'org',
                    'result': None,
                    'properties': properties}}]
            }
        )

    def test_event_not_sent_when_client_has_expired(self, dli_client):
        dli_client._session.has_expired = True

        analytics_handler = AnalyticsHandler(dli_client)
        properties = {
            'package_id': '1',
            'dataset_id': '2',
            'name': 'some_name',
            'dataset_name': 'd_name',
            'datafile_id': 'datafile_id',
            'dictionary_id': 'dictionary_id',
            'api_key': 'key',
            'unknown': 'property'
        }
        properties.pop('unknown')

        analytics_handler.create_event('user', 'org', properties=properties)
        dli_client._session.post.assert_not_called()



import pytest
import keyring


class KeyringTestBackend(keyring.backend.KeyringBackend):
    """A test keyring which always outputs same password
    """
    priority = 1
    servicedict = {}

    def set_password(self, servicename, username, password):
        KeyringTestBackend.servicedict[servicename+username] = password

    def get_password(self, servicename, username):
        return KeyringTestBackend.servicedict.get(servicename+username)

    def delete_password(self, servicename, username):
        try:
            del KeyringTestBackend.servicedict[servicename+username]
        finally:
            pass



def pytest_collection_modifyitems(config, items):
    skip_integration = config.getoption("--skip-integration")
    mark_skip_integration = pytest.mark.skip(reason="skipping integration test")
    for item in items:
        if skip_integration and "integration" in item.keywords:
            item.add_marker(mark_skip_integration)

def pytest_addoption(parser):
    parser.addoption(
        "--skip-integration", action="store_true", default=False,
        help="Skip integration tests."
    )


def pytest_configure(config):
    # set the keyring for keyring lib
    keyring.set_keyring(KeyringTestBackend())



import datetime
import sys
import threading
import time
import webbrowser
from json import JSONDecodeError
import warnings

import keyring
import jwt
import requests
import logging
import socket
from http.cookiejar import CookiePolicy

# import webview
from keyring.errors import PasswordDeleteError, NoKeyringError
from wrapt import ObjectProxy
from http import HTTPStatus
from urllib.parse import urljoin, urlparse, parse_qs, urlencode
from collections import defaultdict
from dli import __version__, __product__
from dli.analytics import AnalyticsHandler
from dli.client.listener import _Listener, can_launch
from dli.client.adapters import DLIBearerAuthAdapter, DLIAccountsV1Adapter, \
    DLIInterfaceV1Adapter, DLISamAdapter
from dli.client.components.urls import sam_urls, identity_urls
from dli.client.environments import _Environment
from dli.siren import siren_to_entity
from dli.client.components.auto_reg_metadata import AutoRegMetadata
from dli.client.components.datafile import Datafile
from dli.client.components.dataset import Dataset, Dictionary
from dli.client.components.me import Me
from dli.client.components.package import Package
from dli.client.components.search import Search
from dli.client.components.accounts import Accounts
from dli.client.exceptions import (
    DatalakeException, InsufficientPrivilegesException,
    InvalidPayloadException, UnAuthorisedAccessException,
    CatalogueEntityNotFoundException, AuthenticationFailure
)
from dli.models.paginator import Paginator
from dli.modules.dataset_module import DatasetModule
from dli.modules.package_module import PackageModule
from dli.models.dictionary_model import DictionaryModel
from dli.models.file_model import FileModel
from dli.models.instance_model import InstanceModel, \
    InstancesCollection as InstancesCollection_
from dli.models.package_model import PackageModel
from dli.models.dataset_model import DatasetModel
from dli.siren import PatchedSirenBuilder

class ModelDescriptor:
    """
    This class is responsible for extending the base type passed
    into the __init__ method with the instance of DliClient it has
    been created within, following the descriptor pattern.

    What this means practicably, is that under _client attribute of the 'new'
    type (class instance) there is a backreference to the DliClient,
    which then permits the type to access the shared session object of DliClient,
    rather than having to pass the session into each instance.

    Using an instance instantiated from the base type will not have the
    _client attribute available.
    """

    def __init__(self, model_cls=None):
        self.model_cls = model_cls

    def __get__(self, instance, owner):
        """Returns a model thats bound to the client instance"""
        return type(
            self.model_cls.__name__, (self.model_cls, ),
            {
                '_client': instance
            }
        )


class DliClient(Accounts, AutoRegMetadata,
                Datafile, Dataset, Dictionary,
                Me, Package, Search):
    """
    Definition of a client. This client mixes in utility functions for
    manipulating packages, datasets and datafiles.
    """

    Dataset = ModelDescriptor(DatasetModel)
    Instance = ModelDescriptor(InstanceModel)
    _InstancesCollection = ModelDescriptor(InstancesCollection_)
    _Pagination = ModelDescriptor(Paginator)
    _File = ModelDescriptor(FileModel)
    _Package = ModelDescriptor(PackageModel)
    _DictionaryV2 = ModelDescriptor(DictionaryModel)

    _packages = ModelDescriptor(PackageModule)
    _datasets = ModelDescriptor(DatasetModule)

    _environment_class = _Environment

    def __init__(self, api_root, host=None, debug=None, strict=True,
                 access_id=None, secret_key=None, use_keyring=True):
        self._environment = self._environment_class(api_root)
        self.access_id = access_id
        self.secret_key = secret_key
        self.host = host
        self.debug = debug
        self.strict = strict
        self.use_keyring = use_keyring
        self.logger = logging.getLogger(__name__)
        self.logger.info(
            'Starting SDK session',
            extra={
               'catalogue': self._environment.catalogue,
               'consumption': self._environment.consumption,
               'strict' : strict,
               'version': __version__
            }
        )

        self._session = self._new_session()
        self._analytics_handler = AnalyticsHandler(self)

        self.packages = self._packages()
        self.datasets = self._datasets()

        if access_id is None and secret_key is not None:
            warnings.warn(
                'The parameter `api_key` will be deprecated in the future. '
                'We will be contacting users in the following months to '
                'explain how to migrate to the new authentication flow.',
                PendingDeprecationWarning
            )


    def _new_session(self):
        session = Session(
            self.access_id,
            self.secret_key,
            self._environment,
            self.host,
            logger=self.logger,
            use_keyring=self.use_keyring
        )
        return session

    @property
    def session(self):
        # if the session expired, then reauth
        # and create a new context
        if self._session.has_expired:
            self._session = self._new_session()
        return self._session


class Session(requests.Session):

    def __init__(
        self, access_id, secret_key, environment, host, auth_key=None,
            logger=None, auth_prompt=True, use_keyring=True
    ):
        super().__init__()
        self.auth_key = None
        self.logger = logger
        self.access_id = access_id
        self.secret_key = secret_key
        self._environment = environment
        self.use_keyring = use_keyring
        self.host = host
        self.siren_builder = PatchedSirenBuilder()
        if auth_prompt:
            self._auth_init(auth_key)
        # mount points to add headers to specific routing requests
        self._set_mount_adapters()

        # Don't allow cookies to be set.
        # The new API will reject requests with both a cookie
        # and a auth header (as there's no predictable crediential to choose).
        #
        # However the old API, once authenticated using a Bearer token, will
        # as a side effect of a request return a oidc_id_token which matches
        # the auth header. This is ignored.
        self.cookies.set_policy(BlockAll())

    def _reload_or_web_flow(self, raises=False):

        check_key = None
        if hasattr(self, "use_keyring") and self.use_keyring:
            try:
                check_key = keyring.get_password(
                    __product__, self._environment.catalogue)
            except NoKeyringError as nke:
                # disable the use of keyring
                self.use_keyring = False
                check_key = None
                print(
                    f"We've noticed that you seem to be trying to run the SDK in a "
                    f"headless machine but your application credential "
                    f" environment variables (DLI_CLIENT_ID and DL_CLIENT_KEY) "
                    f" aren't set or there is no keyring manager available."
                    f" YOUR MAY NEED TO LOGIN EVERY TIME."
                    f"Please contact support at <support email address> or "
                    f"consult {str(nke)}"
                )
                if raises:
                    raise nke

        reloaded = None
        if check_key:
            try:
                reloaded = check_key
                decoded = jwt.decode(str.encode(reloaded), verify=False)
                if decoded.get("exp", 0) <= time.time():
                    try:
                        keyring.delete_password(__product__,
                                                self._environment.catalogue)
                    except PasswordDeleteError as e:
                        if hasattr(self, "logger"):
                            self.logger.info("No such password")
                    finally:
                        reloaded = None
            except Exception as e:
                raise e

        if reloaded:
            if hasattr(self, "logger") and self.logger is not None:
                self.logger.info(f"Using vault JWT "
                                 f"{__product__} "
                                 f"{self._environment.catalogue}")
            self.auth_key = reloaded
        elif not self.secret_key and not self.access_id:
            self.auth_key = self._get_web_auth_key()

    def _auth_init(self, auth_key=None):

        if auth_key:
            self.auth_key = auth_key
        elif self.secret_key and not self.access_id:
            self.auth_key = self._get_auth_key()
        elif self.access_id and self.secret_key:
            self.auth_key = self._get_SAM_auth_key()
        else:
            self._reload_or_web_flow()

        self.decoded_token = self._get_decoded_token()
        self.token_expires_on = self._get_expiration_date()

    def request(self, method, url, *args, **kwargs):

        if not urlparse(url).netloc:
            url = urljoin(self._environment.catalogue, url)

        kwargs.pop('hooks', None)
        hooks = {'response': self._response_hook}

        try:
            if self.logger:
                self.logger.debug(
                    'Request',
                    extra={
                        'method': method,
                        'request': url,
                        '-args': args,
                        '-kwargs': kwargs
                    }
                )

            return super().request(method, url, hooks=hooks, *args, **kwargs)
        except socket.error as e:
            raise ValueError(
                'Unable to process request due to a networking issue '
                'root cause could be a bad connection, '
                'not being on the correct VPN, '
                'or a network timeout '
            ) from e

    @property
    def has_expired(self):
        # We subtract timedelta from the expiration time in order to allow a safety window for
        # a code block to execute after a check has been asserted.
        return datetime.datetime.utcnow() > \
               (self.token_expires_on - datetime.timedelta(minutes=1))

    def _response_hook(self, response, *args, **kwargs):
        # Apologies for the ugly code. The startswith siren check
        # is to make this only relevant to the old API.
        response = Response(response, self.siren_builder)

        if self.logger:
            self.logger.debug(
                'Response',
                extra={
                    # 'content': response.content,
                    'status': response.status_code,
                    'method': response.request.method,
                    'request': response.request.url,
                    'headers': response.request.headers
                }
            )

        if not response.ok:
            exceptions = defaultdict(
                lambda: DatalakeException,
                {HTTPStatus.BAD_REQUEST: InvalidPayloadException,
                 HTTPStatus.UNPROCESSABLE_ENTITY: InvalidPayloadException,
                 HTTPStatus.UNAUTHORIZED: UnAuthorisedAccessException,
                 HTTPStatus.FORBIDDEN: InsufficientPrivilegesException,
                 HTTPStatus.NOT_FOUND: CatalogueEntityNotFoundException}
            )

            try:
                message = response.json()
            except (JSONDecodeError, ValueError):
                message = response.text

            raise exceptions[response.status_code](
                message=message,
                params=parse_qs(urlparse(response.request.url).query),
                response=response
            )

        return response

    def _set_mount_adapters(self):
        self.mount(
            urljoin(self._environment.catalogue, '__api/'),
            DLIInterfaceV1Adapter(self)
        )

        self.mount(
            urljoin(self._environment.catalogue, '__api_v2/'),
            DLIBearerAuthAdapter(self)
        )

        self.mount(
            self._environment.consumption, DLIBearerAuthAdapter(self)
        )

        self.mount(
            urljoin(self._environment.accounts, 'api/identity/v1/'),
            DLIAccountsV1Adapter(self)
        )

        self.mount(
            urljoin(self._environment.accounts, 'api/identity/v2/'),
            DLIBearerAuthAdapter(self)
        )

        self.mount(
            self._environment.sam, DLISamAdapter(self)
        )

        self.mount(
            self._environment.consumption, DLIBearerAuthAdapter(self)
        )

    def _get_decoded_token(self):
        return jwt.decode(self.auth_key, verify=False)

    def _get_expiration_date(self):
        default_timeout = (
            datetime.datetime.utcnow() +
            datetime.timedelta(minutes=55)
        )

        if 'exp' not in self.decoded_token:
            return default_timeout

        return datetime.datetime.utcfromtimestamp(
            self.decoded_token['exp']
        ) - datetime.timedelta(minutes=5)

    def _get_auth_key(self):
        # todo - calling this when JWT from web flow fails.
        try:
            response = self.post(
                '/__api/start-session',
                headers={
                    'Authorization': 'Bearer {}'.format(
                        self.secret_key
                    )
                }
            )
        except DatalakeException as e:
            raise AuthenticationFailure(
                message='Could not authenticate API key'
            ) from e

        return response.text

    def _get_SAM_auth_key(self):

        sam_response = self.post(
            urljoin(self._environment.sam, sam_urls.sam_token),
            data={
                "client_id": self.access_id,
                "client_secret":  self.secret_key,
                "grant_type": "client_credentials"
            },
            hooks={'response': self._response_hook}
        )

        token = sam_response.json()["access_token"]

        catalogue_response = self.post(
            urljoin(self._environment.accounts,
                    identity_urls.identity_token),
            data={
                "client_id": self.access_id,
                "subject_token": token
            },
            hooks={'response': self._response_hook},
        )

        _jwt = catalogue_response.json()["access_token"]
        return _jwt

    def _get_web_auth_key(self, callback=None):
        result, reason = can_launch(_Listener.DEFAULT_PORT)
        if not result:

            if not reason.startswith("[Errno 98]"):
                # we cannot simply raise, since another SDK may be blocking 8080
                # (98 already in use)
                # and be usable - but we can raise if we get a refusal
                # (111 connection refused)
                raise Exception(f"Cannot authenticate via user flow. Contact DL-Support"
                                f" or your system administrator.\nThis occurred because"
                                f" whilst trying to start the callback listener at"
                                f" {_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}.\n"
                                f"We received {reason} from the socket. This may indicate"
                                f" the socket is already in use or your firewall is"
                                f" blocking inbound connections.")
            else:
                # so we need to just hit it first, to see if it replies as expected
                try:
                    query = requests.get(
                        f"{_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}" \
                        f"/get/test")
                    if query.status_code != 404 or \
                            "jwt" not in query.json().keys() or \
                            query.json()["jwt"] != "None":
                        raise Exception(
                            f"Cannot authenticate via user flow. Contact DL-Support"
                            f" or your system administrator.\nThis occurred because"
                            f" whilst trying to contact the callback listener at"
                            f" {_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}.\n"
                            f"We received {reason} from the socket. This may indicate"
                            f" the socket is already in use (and not by another SDK"
                            f" instance or your firewall is blocking inbound "
                            f"connections.")
                    else:
                        result = True
                except requests.ConnectionError as e:
                    # may not even be up
                    result = True

        postbox = _Listener.run(_Listener.DEFAULT_PORT)
        time.sleep(1)
        # we may need a little time to start the instance in
        # some systems - else we request /get before started
        if callback is None:
            url = f"{_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}" \
                  f"/login?postbox={postbox}&{urlencode(self._environment.__dict__)}"
            try:
                # we need to check that the listener can actually listen
                # before we open the browser - so establish that a listener
                # is up and running, or can listen (not firewalled)
                # else we need to tell the user that their firewall is shut.
               webbrowser.open(url, new=1)

            except webbrowser.Error as e:
                print(f"We couldn't open a usable browser for you to authenticate."
                      f"Please copy this {url} into a browser on your system")

            # implement this instead if SAM changes made
            # as we can close the window once value captured.
            # window = webview.create_window(
            #     'Login', f"http://localhost:{_Listener.DEFAULT_PORT}/login",
            #     resizable=False
            # )
            #
            # def get_elements(window):
            #     heading = window.get_elements('#emailaddress')
            #
            #     if not heading:
            #         window = webview.create_window(
            #             'Login', f"https://catalogue-dev.udpmarkit.net/login",
            #             resizable=False
            #         )
            #         webview.start(get_elements, window, debug=False)
            #
            #
            # webview.start(get_elements, window, debug=False)

        else:
            callback(_Listener.DEFAULT_PORT, postbox)

        if result:
            while True:
                query = requests.get(f"{_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}" \
                      f"/get/{postbox}")
                if query.status_code != 200:
                    time.sleep(0.5)
                else:
                    str_jwt = query.json()["jwt"]
                    if str_jwt != "invalid":
                        _jwt = str.encode(str_jwt)
                        break
                    else:
                        print("You have another SDK login open. Complete that first.")
                        sys.exit(0)


            if hasattr(self, "use_keyring") and self.use_keyring:
                keyring.set_password(__product__, self._environment.catalogue, str_jwt)

            # cleanup
            requests.post(f"{_Listener.LOCALHOST}:{_Listener.DEFAULT_PORT}/shutdown")
            # window.destroy()
            return str_jwt


class Response(ObjectProxy):

    def __init__(self, wrapped, builder, *args, **kwargs):
        super(Response, self).__init__(wrapped, *args, **kwargs)
        self.builder = builder

    def to_siren(self):
        # Pypermedias terminology, not mine
        python_object = self.builder._construct_entity(
            self.json()
        ).as_python_object()

        # Keep the response availible
        python_object._raw_response = self

        return python_object

    def to_many_siren(self, relation):
        return [
            siren_to_entity(c) for c in
            self.to_siren().get_entities(rel=relation)
        ]


class BlockAll(CookiePolicy):
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False



from dli.models import AttributesDict


class AccountModel(AttributesDict):
    @classmethod
    def _from_v2_response(cls, data):
        id_ = data.pop('id')
        attributes = data.pop('attributes')
        attributes.pop('ui_settings', None)
        return cls(id=id_, **attributes)


import pytest
import io

from dli.client import builders
from dli.client.exceptions import (
    CatalogueEntityNotFoundException, InvalidPayloadException,
)
from tests.common import SdkIntegrationTestCase

@pytest.mark.integration
class DatasetTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super().setUp()

        self.package_id = self.create_package("test_dataset_functions")
        self.builder = self.dataset_builder(self.package_id, "test_dataset_functions")

    def create_datalake_dataset(self):
        return self.register_s3_dataset(
            package_id=self.package_id,
            dataset_name="test_dataset_functions",
            bucket_name="my-happy-bucket"
        )

    def test_get_unknown_dataset_raises_dataset_not_found_error(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_dataset("unknown")

    def test_can_create_with_data_preview(self):
        self._setup_bucket_and_prefix('test-bucket', 'test')

        csv = io.BytesIO(
            b'a,b,c\n'
            b'1,2,3'
        )

        csv.name = 'file/test.csv'

        builder = self.dataset_builder(
            package_id=self.package_id,
            dataset_name='datapreview-test',
            sample_data=csv,
            data_preview_type='STATIC'
        ).with_external_s3_storage(
            bucket_name='test-bucket',
            aws_account_id=self.aws_account_id,
            prefix='test'
        )

        dataset = self.client.register_dataset(builder)

        assert dataset.data_preview_file_name == 'test.csv'
        with dataset.sample_data.file() as f:
            assert f.read() == (
                b'a,b,c\n'
                b'1,2,3'
            )

    def test_builder_sample_data_validation(self):
        assert builders.DatasetBuilder(
            sample_data='1234',
            data_preview_type='STATIC'
        )

        assert builders.DatasetBuilder(
            sample_data='1234',
        )._data['sample_data']

        with pytest.raises(ValueError):
            assert builders.DatasetBuilder(
                sample_data='1234',
                data_preview_type='LIVE'
            )

    def test_can_edit_with_data_preview(self):
        self._setup_bucket_and_prefix('test-bucket', 'test')

        csv = io.BytesIO(
            b'a,b,c\n'
            b'1,2,3'
        )

        csv.name = 'test.csv'

        builder = self.dataset_builder(
            package_id=self.package_id,
            dataset_name='datapreview-test',
            sample_data=csv,
            data_preview_type='STATIC'
        ).with_external_s3_storage(
            bucket_name="test-bucket",
            aws_account_id=self.aws_account_id,
            prefix="test"
        )

        dataset = self.client.register_dataset(builder)

        csv = io.BytesIO(
            b'different,stuff\n'
            b'more,file'
        )
        csv.name = 'new_test.csv'

        dataset = self.client.edit_dataset(
            dataset.dataset_id,
            sample_data=csv,
        )

        with dataset.sample_data.file() as f:
            assert f.read() == (
                b'different,stuff\n'
                b'more,file'
            )

    def test_can_get_dataset_by_short_code(self):
        builder = self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"
        )
        builder._data.update({'short_code': 'testGetByShortCode'})
        expected = self.client.register_dataset(builder)

        dataset_by_short_code = self.client.get_dataset(
            dataset_short_code='testGetByShortCode'
        )
        self.assertEqual(expected.dataset_id, dataset_by_short_code.dataset_id)

    def test_get_dataset_by_short_code_returns_404_if_organisation_does_not_match(self):
        self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"
        )

        with pytest.raises(CatalogueEntityNotFoundException):
            self.client.get_dataset(
                dataset_short_code='testdatasetfunctions', organisation_short_code='abc'
            )

    def test_can_get_dataset_by_id_or_name(self):
        builder = self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"
        )
        package = self.client.get_package(id=self.package_id)
        expected = self.client.register_dataset(builder)
        actual = self.client.get_dataset(expected.dataset_id)

        self.assertEqual(expected.dataset_id, actual.dataset_id)

        # By id
        dataset_by_id = self.client.get_dataset(id=expected.dataset_id)
        self.assertEqual(expected.dataset_id, dataset_by_id.dataset_id)

        # By name and package id
        dataset_by_name_and_package_id = self.client.get_dataset(name=expected.name, package_id=package.package_id)
        self.assertEqual(expected.dataset_id, dataset_by_name_and_package_id.dataset_id)

        # By name and package name
        dataset_by_name_and_package_name = self.client.get_dataset(name=expected.name, package_name=package.name)
        self.assertEqual(expected.dataset_id, dataset_by_name_and_package_name.dataset_id)

        # By name and package name and package id
        dataset_by_name_and_package_name_and_id = self.client.get_dataset(name=expected.name,
                                                                          package_id=package.package_id,
                                                                          package_name=package.name)
        self.assertEqual(expected.dataset_id, dataset_by_name_and_package_name_and_id.dataset_id)

    def test_cannot_get_dataset_by_name_if_package_id_and_package_name_mismatch(self):
        builder = self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"
        )
        package = self.client.get_package(id=self.package_id)
        dataset = self.client.register_dataset(builder)

        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_dataset(name=dataset.name, package_id=package.package_id, package_name="unknown")
            self.client.get_dataset(name=dataset.name, package_id="unknown", package_name=package.name)

    def test_retrieve_keys_for_unknown_dataset_raises_error(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_s3_access_keys_for_dataset("unknown")

    def test_can_retrieve_keys_for_single_dataset(self):
        dataset = self.create_datalake_dataset()
        keys = self.client.get_s3_access_keys_for_dataset(dataset.dataset_id)

        self.assertIsNotNone(keys.access_key_id)
        self.assertIsNotNone(keys.session_token)
        self.assertIsNotNone(keys.secret_access_key)

    def test_can_retrieve_keys_for_multiple_datasets(self):
        dataset1 = self.register_s3_dataset(
            package_id=self.package_id, dataset_name="dataset1", bucket_name="my-happy-bucket-1"
        )
        dataset2 = self.register_s3_dataset(
            package_id=self.package_id, dataset_name="dataset2", bucket_name="my-happy-bucket-2"
        )

        keys = self.client.get_s3_access_keys_for_dataset(
            dataset1.dataset_id,
            dataset2.dataset_id
        )

        self.assertIsNotNone(keys.access_key_id)
        self.assertIsNotNone(keys.session_token)
        self.assertIsNotNone(keys.secret_access_key)

    def test_can_not_create_dataset_without_location(self):
        with self.assertRaises(Exception):
            self.client.register_dataset(self.builder)

    def test_can_create_dataset_with_other_location(self):
        builder = self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"
        )
        dataset = self.client.register_dataset(builder)

        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.description, self.builder.payload["description"])
        self.assertEqual(dataset.location.type, "Other")

    def test_can_create_dataset_with_short_code(self):
        builder = self.dataset_builder(
            package_id=self.package_id,
            dataset_name='datapreview-test',
            short_code='someRandomShortCode'
        ).with_external_storage(
            location='jdbc://connectionstring:1232/my-db'
        )
        dataset = self.client.register_dataset(builder)

        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.short_code, builder._data['short_code'])

    def test_can_create_dataset_with_external_bucket(self):
        dataset = self.register_s3_dataset(self.package_id, "test_dataset", "my-happy-external-bucket")

        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.location.type, "S3")
        self.assertEqual(dataset.location.bucket, "my-happy-external-bucket")

    def test_can_edit_dataset_with_same_values(self):
        dataset = self.create_datalake_dataset()

        updated = self.client.edit_dataset(
            dataset.dataset_id,
            name=dataset.name,
            description=dataset.description,
            content_type=dataset.content_type,
            data_format=dataset.data_format
        )

        self.assertEqual(dataset.dataset_id, updated.dataset_id)
        self.assertEqual(dataset.package_id, updated.package_id)
        self.assertEqual(dataset.name, updated.name)
        self.assertEqual(dataset.location, updated.location)
        self.assertEqual(dataset.created_at, updated.created_at)
        # updated has changed
        self.assertNotEqual(dataset.updated_at, updated.updated_at)

    def test_can_edit_and_change_values(self):
        dataset = self.create_datalake_dataset()
        # change location to other
        builder = builders.DatasetLocationBuilder().with_external_storage("changing-location")

        updated = self.client.edit_dataset(
            dataset.dataset_id,
            location_builder=builder,
            description="new desc",
            content_type="content type",
            keywords=["test", "2"],
            load_type="Full Load"
        )

        self.assertEqual(updated.dataset_id, dataset.dataset_id)
        self.assertEqual(updated.package_id, dataset.package_id)
        self.assertEqual(updated.name, dataset.name)
        self.assertEqual(updated.description, "new desc")
        self.assertEqual(updated.content_type, "content type")
        self.assertEqual(updated.keywords, ["test", "2"])
        self.assertEqual(updated.location.type, "Other")
        self.assertEqual(updated.load_type, "Full Load")

    def test_can_create_dataset_with_load_type(self):
        dataset = self.client.register_dataset(builders.DatasetBuilder(
            package_id=self.package_id,
            name="dataset_with_load_type",
            description="a testing dataset",
            content_type="Pricing",
            data_format="CSV",
            publishing_frequency="Daily",
            taxonomy=[],
            load_type="Incremental Load",
            data_preview_type="NONE"
        ).with_external_storage("test")
                                               )

        self.assertEqual(dataset.load_type, "Incremental Load")

    def test_can_delete_dataset(self):
        dataset = self.create_datalake_dataset()
        self.client.delete_dataset(dataset.dataset_id)

        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_dataset(dataset.dataset_id)

    def test_delete_unknown_dataset_raises_error(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_dataset("unknown")

    def test_register_dataset_in_unknown_package_raises_error(self):
        builder = builders.DatasetBuilder(
            package_id="unknown",
            name="test",
            description="a testing dataset",
            content_type="Pricing",
            data_format="CSV",
            publishing_frequency="Daily",
            taxonomy=[],
            data_preview_type="NONE"
        )
        builder = builder.with_external_storage("s3://my-happy-bucket")

        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.register_dataset(builder)

@pytest.mark.integration
class DatasetDatafilesTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super().setUp()

    def test_get_dataset_datafiles_raises_exception_if_dataset_does_not_exists(self):
        with self.assertRaises(Exception):
            self.client.get_datafiles("unknown")

    def test_get_dataset_datafiles_returns_empty_when_no_datafiles(self):
        package_id = self.create_package(
            name="test_get_dataset_datafiles_returns_empty_when_no_datafiles"
        )
        dataset = self.client.register_dataset(
            self.dataset_builder(
                package_id,
                "test_get_dataset_datafiles_returns_empty_when_no_datafiles"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )
        datafiles = self.client.get_datafiles(dataset.dataset_id)
        self.assertEqual(datafiles, [])

    def test_get_dataset_datafiles_returns_datafiles_for_dataset(self):
        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        package_id = self.create_package(
            name="test_get_dataset_datafiles_returns_datafiles_for_dataset"
        )
        dataset = self.client.register_dataset(
            self.dataset_builder(
                package_id,
                "test_get_dataset_datafiles_returns_datafiles_for_dataset"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )

        for i in range(1, 4):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "datafile %s" % i,
                files,
                data_as_of="2018-10-1%s" % i
            )

        datafiles = self.client.get_datafiles(dataset.dataset_id)
        self.assertEqual(len(datafiles), 3)

        datafiles_paged = self.client.get_datafiles(dataset.dataset_id, count=2)
        self.assertEqual(len(datafiles_paged), 2)

        # Other Lookup scenarios
        datafiles_search_by_as_of_date = self.client.get_datafiles(dataset.dataset_id, as_of_date_start='2018-10-12')
        self.assertEqual(len(datafiles_search_by_as_of_date), 2)
        self.assertTrue(all(df.name in ["datafile 2", "datafile 3"] for df in datafiles_search_by_as_of_date))

        datafiles_search_by_as_of_date = self.client.get_datafiles(dataset.dataset_id, as_of_date_end='2018-10-12')
        self.assertEqual(len(datafiles_search_by_as_of_date), 2)
        self.assertTrue(all(df.name in ["datafile 1", "datafile 2"] for df in datafiles_search_by_as_of_date))

        datafiles_search_by_as_of_date = self.client.get_datafiles(dataset.dataset_id, as_of_date_start='2018-10-10',
                                                                   as_of_date_end='2018-10-11')
        self.assertEqual(len(datafiles_search_by_as_of_date), 1)
        self.assertTrue(all(df.name == "datafile 1" for df in datafiles_search_by_as_of_date))

        datafiles_search_by_as_of_date = self.client.get_datafiles(dataset.dataset_id, name_contains='datafile 2',
                                                                   as_of_date_start='2018-10-10',
                                                                   as_of_date_end='2018-10-13')
        self.assertEqual(len(datafiles_search_by_as_of_date), 1)
        self.assertTrue(all(df.name == "datafile 2" for df in datafiles_search_by_as_of_date))

    def test_get_dataset_datafiles_raises_exception_for_invalid_as_of_date_params(self):
        package_id = self.create_package(
            name="test_get_dataset_datafiles_raises_exception_for_invalid_as_of_date_params"
        )
        dataset = self.client.register_dataset(
            self.dataset_builder(
                package_id,
                "test_get_dataset_datafiles_raises_exception_for_invalid_as_of_date_params"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )

        with self.assertRaises(InvalidPayloadException):
            self.client.get_datafiles(dataset.dataset_id, as_of_date_end='blah blah')
            self.client.get_datafiles(dataset.dataset_id, as_of_date_start='blah blah')
            self.client.get_datafiles(dataset.dataset_id, as_of_date_start='11/12/2017', as_of_date_end='15/12/2017')

    def test_get_dataset_datafiles_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(
            lambda c: self.client.get_datafiles("some_dataset", count=c))

    def test_get_latest_datafile_raises_exception_if_dataset_does_not_exists(self):
        with self.assertRaises(Exception):
            self.client.get_latest_datafile("unknown")

    def test_get_latest_datafile_raises_exception_if_dataset_has_no_datafiles(self):
        package_id = self.create_package(
            name="test_get_latest_datafile_raises_exception_if_dataset_has_no_datafiles"
        )
        dataset = self.client.register_dataset(
            self.dataset_builder(
                package_id,
                "test_get_latest_datafile_raises_exception_if_dataset_has_no_datafiles"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )

        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        for i in range(4, 0, -1):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "datafile %s" % i,
                files,
                data_as_of="2018-10-1%s" % i
            )

        latest_datafile = self.client.get_latest_datafile(dataset.dataset_id)

        self.assertEqual(latest_datafile.name, 'datafile 4')

    def test_get_latest_datafile_returns_latest_datafile(self):
        package_id = self.create_package(
            name="test_get_latest_datafile_returns_latest_datafile"
        )
        dataset = self.client.register_dataset(
            self.dataset_builder(
                package_id,
                "test_get_latest_datafile_returns_latest_datafile"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )

        with self.assertRaises(Exception):
            self.client.get_latest_datafile(dataset.dataset_id)


@pytest.mark.integration
class DatasetSchemasTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super().setUp()

        dataset = self.register_s3_dataset(
            package_id=self.create_package("test_dataset_schema_functions"),
            dataset_name="test_dataset_schema_functions",
            bucket_name="my-happy-bucket"
        )
        self.dataset_id = dataset.dataset_id

    def create_schema_payload(self, version, valid_as_of, **kwargs):
        payload = {
            'version': version,
            'valid_as_of': valid_as_of,
            'fields': [
                {
                    'name': 'field1',
                    'type': 'Int',
                    'nullable': True
                },
            ]
        }

        payload.update(**kwargs)
        return payload

    def test_cannot_register_schema_for_missing_or_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.register_schema(
                'unknown', version='1a', valid_as_of='2018-10-31', fields=[
                    {
                        'name': 'field2',
                        'nullable': False,
                        'type': 'Int',
                        'metadata': {
                            'random': 'keys'
                        }
                    },
                ])

    def test_cannot_register_schema_with_invalid_payload(self):
        my_schema_fields = [
            {
                'name': 'field1',
                'type': 'Int',
                'nullable': True
            },
        ]
        # fields cannot be empty or contain invalid field definitions
        with self.assertRaises(InvalidPayloadException):
            self.client.register_schema(self.dataset_id, version='1a', valid_as_of='2018-10-31', fields=None)
        with self.assertRaises(InvalidPayloadException):
            self.client.register_schema(self.dataset_id, version='1a', valid_as_of='2018-10-31',
                                        fields=[{'random': 'keys'}])

        # valid_as_of should be date in ISO format
        with self.assertRaises(InvalidPayloadException):
            self.client.register_schema(self.dataset_id, version='1a', valid_as_of='31/10/2018',
                                        fields=my_schema_fields)
        with self.assertRaises(InvalidPayloadException):
            self.client.register_schema(self.dataset_id, version='1a', valid_as_of='halloween', fields=my_schema_fields)

        # partitions if provided can be empty, but must not contain invalid key value pairs
        self.client.register_schema(self.dataset_id, version='1a', valid_as_of='2018-10-31',
                                    fields=my_schema_fields, partitions=[])
        with self.assertRaises(InvalidPayloadException):
            self.client.register_schema(self.dataset_id, version='1a', valid_as_of='2018-10-31',
                                        fields=my_schema_fields, partitions=[{'random': 'keys'}])

    def test_can_register_schema(self):
        my_schema_fields = [
            {
                'name': 'field1',
                'type': 'Int',
                'nullable': True
            },
            {
                'name': 'field2',
                'type': 'Int',
                'nullable': False,
                'metadata': {
                    'random': 'keys'
                }
            },
        ]
        my_schema_partitions = [
            {
                'name': 'Who goes there?',
                'type': 'Phantom'
            }
        ]
        params = {
            'version': '1a',
            'valid_as_of': '2018-10-31',
            'fields': my_schema_fields,
            'partitions': my_schema_partitions,
            'description': 'My Funky Schema',
        }

        schema = self.client.register_schema(self.dataset_id, **params)
        self.assertIsNotNone(schema)
        self.assertEqual(schema.version, params['version'])
        self.assertEqual(schema.valid_as_of, params['valid_as_of'])
        self.assertEqual(len(schema.fields), 2)
        self.assertTrue(all([f['name'] in ['field1', 'field2'] for f in schema.fields]))
        self.assertEqual(len(schema.partitions), 1)

    def test_can_register_spark_schema(self):
        spark_schema = {
            'fields': [
                {
                    'metadata': {},
                    'name': 'field_1',
                    'nullable': True,
                    'type': 'string'
                },
                {
                    'metadata': {},
                    'name': 'field_2',
                    'nullable': True,
                    'type': 'long'
                }
            ],
        }
        schema = self.client.register_spark_schema(
            self.dataset_id,
            version='1a',
            valid_as_of='2018-10-31',
            spark_schema_json=spark_schema
        )
        self.assertIsNotNone(schema)
        self.assertEqual(schema.version, '1a')
        self.assertEqual(schema.valid_as_of, '2018-10-31')
        self.assertEqual(len(schema.fields), 2)
        self.assertTrue(all([f['name'] in ['field_1', 'field_2'] for f in schema.fields]))

    def test_cannot_get_schema_for_missing_or_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schema(None)
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schema('unknown')

    def test_cannot_get_schema_for_dataset_with_no_schemas(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schema(self.dataset_id)

    def test_can_get_schema_by_version(self):
        payload = self.create_schema_payload('1a', '2018-10-31')
        self.client.register_schema(self.dataset_id, **payload)
        schema = self.client.get_schema(self.dataset_id, payload['version'])
        self.assertIsNotNone(schema)
        self.assertEqual(schema.version, payload['version'])
        assert schema.fields

    def test_can_get_schema_without_version_returns_the_latest(self):
        payload_a = self.create_schema_payload('1a', '2018-10-31')
        payload_b = self.create_schema_payload('1b', '2018-11-05')

        self.client.register_schema(self.dataset_id, **payload_a)
        self.client.register_schema(self.dataset_id, **payload_b)

        schema = self.client.get_schema(self.dataset_id)
        self.assertIsNotNone(schema)
        self.assertEqual(schema.version, payload_b['version'])

    def test_cannot_get_schemas_for_missing_or_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schemas(None)
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schemas('unknown')

    def test_get_schemas_returns_empty_for_dataset_with_no_schemas(self):
        self.assertEqual(self.client.get_schemas(self.dataset_id), [])

    def test_can_get_schemas_for_dataset_and_schemas_are_sorted_by_valid_as_of_date(self):
        payload_a = self.create_schema_payload('1a', '2018-10-31')
        payload_b = self.create_schema_payload('1b', '2018-11-05')
        payload_c = self.create_schema_payload('1c', '2018-11-02')

        self.client.register_schema(self.dataset_id, **payload_a)
        self.client.register_schema(self.dataset_id, **payload_b)
        self.client.register_schema(self.dataset_id, **payload_c)

        all_schemas = self.client.get_schemas(self.dataset_id)
        self.assertEqual(len(all_schemas), 3)
        self.assertEqual(all_schemas[0].version, payload_b['version'])

        fewer_schemas = self.client.get_schemas(self.dataset_id, count=2)
        self.assertEqual(len(fewer_schemas), 2)
        self.assertEqual(fewer_schemas[0].version, payload_b['version'])

    def test_cannot_delete_schema_for_missing_or_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_schema(None, '1a')

        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_schema('unknown', '1a')

    def test_cannot_delete_schema_if_missing_or_unknown_version(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_schema(self.dataset_id, None)
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_schema(self.dataset_id, 'unknown')

    def test_can_delete_schema_by_version(self):
        # Create schema first
        payload = self.create_schema_payload('1a', '2018-10-31')
        self.client.register_schema(self.dataset_id, **payload)
        self.assertIsNotNone(self.client.get_schema(self.dataset_id, payload['version']))
        # Now Delete
        self.client.delete_schema(self.dataset_id, payload['version'])
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schema(self.dataset_id, payload['version'])

    def test_cannot_edit_schema_for_missing_or_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_schema(None, '1a')
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_schema('unknown', '1a')

    def test_cannot_edit_schema_if_missing_or_unknown_version(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_schema(self.dataset_id, None)
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_schema(self.dataset_id, 'unknown')

    def test_can_edit_schema_by_version(self):
        # Create schema first
        payload = self.create_schema_payload('1a', '2018-10-31')
        self.client.register_schema(self.dataset_id, **payload)
        self.assertIsNotNone(self.client.get_schema(self.dataset_id, payload['version']))
        # Now Update valid as of
        result = self.client.edit_schema(self.dataset_id, payload['version'], valid_as_of='2018-11-05')

        self.assertEqual(result.valid_as_of, '2018-11-05')
        # Can update version too if available
        result = self.client.edit_schema(self.dataset_id, payload['version'], new_version='1b')
        self.assertEqual(result.version, '1b')
        # version now updated
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_schema(self.dataset_id, payload['version'])




import os

import pytest
import tempfile

from tests.common import SdkIntegrationTestCase
from unittest.mock import patch


@pytest.mark.integration
class SDKSamples(SdkIntegrationTestCase):
    """
    This spec is just meant to catch issues in terms of signature changes
    all the methods are included in one way or another in the SDK documentation
    as samples.

    Avoid including assertions or similar as these tests are not replacements
    for unit or integration tests on these functions.
    """

    def test_samples_get_datasets_in_package(self):
        client = self.client
        package_id = self.create_package("test_get_datasets_in_package")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_dataset_details").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)

        # doc-start
        # returns list of datasets ordered by ascending name
        datasets = client.get_package_datasets(package_id)

        print("Retrieved {} datasets.".format(len(datasets)))
        for ds in datasets:
            print(ds.dataset_id, ds.name)

    def test_get_dataset_details(self):
        client = self.client
        package_id = self.create_package("package_for_test_get_dataset_details")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_dataset_details").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        # doc-start
        # given a `dataset_id`, dataset metadata can be retrieved as
        dataset = client.get_dataset(dataset_id)

    def test_get_datafiles_in_dataset(self):
        client = self.client
        package_id = self.create_package("package_for_test_get_datafiles_in_dataset")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_datafiles_in_dataset").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        datafile = client.register_datafile_metadata(
            dataset_id,
            "datafile_for_test_get_datafiles_in_dataset",
            files,
            '2000-01-01'
        )

        # doc-start
        # returns list of datafiles ordered by descending creation date
        datafiles = client.get_datafiles(dataset_id)

        print("Retrieved {} datafiles.".format(len(datafiles)))
        for df in datafiles:
            print((df.datafile_id, df.files))

    def test_download_datafile(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            client = self.client
            package_id = self.create_package("package_for_test_download_datafile")
            dataset = self.register_s3_dataset(package_id, "dataset_for_test_download_datafile", "external-bucket", prefix="samples")
            dataset_id = dataset.dataset_id

            datafile = client.register_s3_datafile(
                dataset_id,
                "datafile_for_test_download_datafile",
                [
                    '../test_sandbox/samples/data/AAPL.csv',
                    '../test_sandbox/samples/data/MSFT.csv'
                ],
                "samples/",
                '2000-01-01'
            )
            datafile_id = datafile.datafile_id

            with patch('dli.client.s3.Client.download_files_from_s3_path') \
                    as s3_download:
                with tempfile.TemporaryDirectory() as destination:
                    # doc-start
                    client.download_datafile(
                        datafile_id,
                        destination,
                        flatten=False
                    )

                    import os
                    os.listdir(destination)  # shows: ['AAPL.csv', 'MSFT.csv']

    def test_create_datafile_in_dl_s3_storage(self):
        with patch('dli.client.s3.Client.upload_files_to_s3',
                   self.patch_upload_files_to_s3) as s3_upload:
            client = self.client

            package_id = self.create_package("package_for_test_create_datafile_in_dl_s3_storage")
            dataset = self.register_s3_dataset(package_id, "dataset_for_test_create_datafile_in_dl_s3_storage", "external-bucket")
            dataset_id = dataset.dataset_id

            # doc-start
            # given a dataset with data lake managed storage
            # we can register and upload files into S3
            # which then can be used by your consumers
            datafile = client.register_s3_datafile(
                dataset_id,
                "My Datafile",
                files=[
                    '../test_sandbox/samples/data/AAPL.csv',
                    '../test_sandbox/samples/data/MSFT.csv'
                ],
                s3_prefix="prefix",
                data_as_of='2000-01-01'
            )

    def test_create_datafile_in_external_s3_storage(self):
        client = self.client
        package_id = self.create_package("package_for_test_create_datafile_in_external_s3_storage")
        dataset = self.register_s3_dataset(package_id, "dataset_for_test_create_datafile_in_external_s3_storage", "external-bucket", prefix="path")
        dataset_id = dataset.dataset_id

        # doc-start
        # given a dataset with external storage, we want to register a new datafile under it
        # in this case the bucket `external-bucket` is not managed by the data lake, so the upload
        # needs to be done manually
        datafile = client.register_datafile_metadata(
            dataset_id,
            name="test",
            files=[
                {"path": "s3://external-bucket/path/to/file/A", "size": 10000},
                {"path": "s3://external-bucket/path/to/file/B", "size": 15222}
            ],
            data_as_of='2000-01-01'
        )

    def test_update_and_delete_datafile(self):
        client = self.client
        package_id = self.create_package("package_for_test_update_and_delete_datafile")
        builder = self.dataset_builder(package_id, "dataset_for_test_update_and_delete_datafile").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        datafile = client.register_datafile_metadata(
            dataset_id,
            "datafile_for_test_update_and_delete_datafile",
            files,
            '2000-01-01'
        )

        # doc-start
        # now that we have created a datafile, and have an id (under `datafile.datafile_id`)
        # we can make changes to the metadata
        # this function for example, would change the name
        # while leaving all other attributes as they are.
        updated = client.edit_datafile_metadata(
            datafile.datafile_id,
            name="Updating Datafile."
        )

        # we can also mark the datafile as deleted
        # if we don't want it to be available for consumption anymore.
        # this call only deletes the metadata and leaves
        # the actual data intact.
        client.delete_datafile(updated.datafile_id)



import contextlib
import inspect
import warnings
from collections import UserDict
from typing import Mapping

from dli.client.components.urls import dataset_urls


def log_public_functions_calls_using(decorators, class_fields_to_log=None):
    if not class_fields_to_log:
        class_fields_to_log = []

    def decorate(cls):
        functions_to_exclude = inspect.getmembers(AttributesDict, inspect.isfunction)
        functions_to_decorate = [
            func for func in inspect.getmembers(cls, inspect.isfunction)
            if func not in functions_to_exclude and not func[0].startswith('__')
        ]
        for function_meta in functions_to_decorate:
            function_name = function_meta[0]
            for decorator in decorators:
                setattr(
                    cls,
                    function_name,
                    decorator(getattr(cls, function_name),
                              class_fields_to_include=class_fields_to_log)
                )
        return cls
    return decorate


class SampleData:
    def __init__(self, parent):
        self._parent = parent
        self._client = parent._client

    def schema(self):
        """
        Returns the schema data and first rows of sample data.

        :returns: attributes dictionary
        """
        response = self._client.session.get(
            dataset_urls.v2_sample_data_schema.format(id=self._parent.id)
        )

        return AttributesDict(**response.json()['data']['attributes'])

    @contextlib.contextmanager
    def file(self):
        """
        Provides a file like object containing sample data.

        Example usage:

        .. code-block:: python

            dataset = self.client.get_dataset(dataset_id)
            with dataset.sample_data.file() as f:
                dataframe = pandas.read_csv(f)
        """
        response = self._client.session.get(
            dataset_urls.v2_sample_data_file.format(id=self._parent.id),
            stream=True
        )
        # otherwise you get raw secure
        response.raw.decode_content = True
        yield response.raw
        response.close()


class AttributesDict(UserDict):

    def __init__(self, *args, **kwargs):
        # recursively provide the rather silly attribute
        # access
        data = {}

        for arg in args:
            data.update(arg)

        data.update(**kwargs)

        for key, value in data.items():
            if isinstance(value, Mapping):
                self.__dict__[key] = AttributesDict(value)
            else:
                self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def _asdict(self, *args, **kwargs):
        warnings.warn(
            'This method is deprecated as it returns itself.',
            DeprecationWarning
        )

        return self

    def __repr__(self):
        attributes = ' '.join([
            '{}={}'.format(k, v) for k,v in self.__dict__.items()
            if not k.startswith('_')
        ])

        return "{}({})".format(self.__class__.__name__, attributes)



import textwrap
import warnings
from collections import OrderedDict
from typing import Dict

from dli.models.paginator import Paginator
from dli.client.components.urls import package_urls
from dli.models import AttributesDict
from dli.models.dataset_model import DatasetModel


class PackageModel(AttributesDict):
    """
    Represents a package pulled back from Me consumables
    """

    @classmethod
    def _raw_dict(cls, v1_response):
        return v1_response['properties']

    @classmethod
    def _from_v1_response_to_v2(cls, v1_response):
        response = cls._client.session.get(
            package_urls.v2_package_index.format(
                id=v1_response['properties']['packageId']
            )
        )

        return cls(response.json())

    def __init__(self, json):
        super().__init__(id=json['data']['id'], **json['data']['attributes'])
        self._paginator = Paginator(
            package_urls.v2_package_datasets.format(id=self.id),
            self._client.Dataset,
            self._client.Dataset._from_v2_response_unsheathed
        )
        self.__shape = None

    @property
    def shape(self):
        """
        :returns: Count the number of datasets.
        :rtype: Int
        """
        if self.__shape is None:
            self.__shape = len(self.datasets())
        return self.__shape

    def datasets(self) -> Dict[str, DatasetModel]:
        """
        :returns: Dictionary of datasets in a package.
        :rtype: OrderedDict[id: str, DatasetModel]
        """
        return OrderedDict([
            (v.name, v) for v in self._paginator
        ])

    def contents(self):
        """Print information about all the datasets in this package."""
        for p in self.datasets().items():
            print(str(p[1]))

    def __str__(self):
        separator = "-"*80
        split_description = "\n".join(textwrap.wrap(self.description, 80))
        split_keywords = "\n".join(self.keywords or [])
        split_documentation = "\n".join(textwrap.wrap(self.documentation, 80))

        # When the fields are re-named in the JSON, it breaks our mapping. Use
        #   return str(self.__dict__)
        # to get the new field names then update the below.
        return f"\nPACKAGE \"{self.name}\" " \
               f"(Contains: {self.shape} datasets)\n" \
               f">> ID: {self.id} \n" \
               f">> Sensitivity: {self.data_sensitivity} " \
               f"/ Accessible: {self.has_access}\n" \
               f"\n" \
               f"{split_description}\n" \
               f"Documentation: {split_documentation}\n\n" \
               f"Keywords:\n{split_keywords}\n"\
               f"{separator}"

    def __repr__(self):
        return f'<Package short_code={self.name}>'



import pytest

from tests.common import SdkIntegrationTestCase


@pytest.mark.integration
class MeTestCase(SdkIntegrationTestCase):

    def assert_my_entities_are_returned(self, func):
        assert len(func()) > 0
        assert len(func(count=1)) == 1

    def test_get_my_consumed_packages_validates_page_size(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(
            lambda c: self.client.get_my_consumed_packages(count=c)
        )

    def test_my_consumed_packages_returns_my_consumed_packages(self):
        package_id = self.create_package('test_consumed_packages', access="Unrestricted")
        package = self.client.get_package.__self__._get_package(package_id=package_id)

        self.client.session.put(
            f'/__api/packages/{package_id}/access/',
            json={
                'accountId': 'iboxx',
                'operation': 'request',
                'comment': 'Hello there!'
            }
        )

        self.assert_my_entities_are_returned(self.client.get_my_consumed_packages)

    def test_get_my_packages_validates_page_size(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(
            lambda c: self.client.get_my_packages(count=c)
        )

    def test_get_my_packages_returns_packages(self):
        self.create_package("test_me_functions")
        self.assert_my_entities_are_returned(self.client.get_my_packages)



import pytest

from unittest.mock import MagicMock
from unittest.mock import call

from dli.client.dli_client import DliClient


class Client(DliClient):

    def __init__(self, strict=True):
        self.logger = MagicMock()
        self._session = MagicMock()
        self.strict = strict

    @property
    def session(self):
        return self._session

    def log_debug(self):
        self.logger.debug('some debug message')

    def log_info(self):
        self.logger.info('some info message')

    def log_warning(self):
        self.logger.warning('some warning message')

    def log_error(self):
        self.logger.error('some error message')

    def log_exception(self):
        raise Exception('some exception message')


@pytest.fixture
def client():
    yield Client(strict=True)


@pytest.fixture
def client_not_strict_exception_logging():
    yield Client(strict=False)


class TestComponentsAspectWrapper:

    def test_debug_logging_is_logged_correctly(self, client):
        client.log_debug()

        assert client.logger.info.call_count == 1, 'One call at info level to log details of the function.'
        assert client.logger.debug.call_args_list == [call('some debug message')]
        assert client.logger.warning.call_count == 0
        assert client.logger.error.call_count == 0

    def test_info_logging_is_logged_correctly(self, client):
        client.log_info()

        assert client.logger.debug.call_count == 0
        assert client.logger.info.call_count == 2
        assert client.logger.info.call_args_list[1] == call('some info message')
        assert client.logger.warning.call_count == 0
        assert client.logger.error.call_count == 0

    def test_warning_logging_is_logged_correctly(self, client):
        client.log_warning()

        assert client.logger.info.call_count == 1, 'One call at info level to log details of the function.'
        assert client.logger.debug.call_count == 0
        assert client.logger.warning.call_args_list == [call('some warning message')]
        assert client.logger.error.call_count == 0

    def test_error_logging_is_logged_correctly(self, client):
        client.log_error()

        assert client.logger.info.call_count == 1, 'One call at info level to log details of the function.'
        assert client.logger.debug.call_count == 0
        assert client.logger.warning.call_count == 0
        assert client.logger.error.call_args_list == [call('some error message')]

    def test_exception_logging_is_logged_correctly_when_strict(self, client):
        with pytest.raises(Exception) as excinfo:
            client.log_exception()
            assert 'some exception message' in str(excinfo.value)

        assert client.logger.info.call_count == 1
        assert client.logger.debug.call_count == 0
        assert client.logger.warning.call_count == 0
        assert client.logger.error.call_count == 0
        assert client.logger.exception.call_count == 1, 'Use logger exception to capture the stack trace.'
        message, args = client.logger.exception.call_args_list[0]
        assert message == ('Unhandled Exception',)
        assert 'locals' in args['extra'], 'Local variables should be added to the message'
        assert args['stack_info'] is False
        assert len(args['extra']['locals']) > 0, 'The local variables must not be empty'
        # assert client.logger.exception.call_args_list == [call('Unhandled Exception', stack_info=True)]

    def test_exception_logging_is_logged_correctly_when_not_strict(self, client_not_strict_exception_logging):
        with pytest.raises(Exception) as excinfo:
            client_not_strict_exception_logging.log_exception()
            assert 'some exception message' in str(excinfo.value)

        logger = client_not_strict_exception_logging.logger
        assert logger.info.call_count == 1
        assert logger.debug.call_count == 0
        assert logger.warning.call_count == 1, \
            'Must display warning that exception message and stack trace ' \
            'have been suppressed'
        assert logger.error.call_count == 0
        assert logger.exception.call_count == 0, \
            'Must be surpressing logger exception and stack trace.'
        assert logger.warning.call_count == 1, \
            'Warning printed instead of logging exception + stack trace'
        message, _ = logger.warning.call_args_list[0]
        assert 'An exception occurred so we are returning None.' in message[0]



import warnings

from dli.client.builders import DatasetBuilder
from dli.client.components import SirenComponent
from dli.client.utils import ensure_count_is_valid
from dli.models.dictionary_model import DictionaryModel
from dli.siren import siren_to_entity

from dli.client.components.urls import dataset_urls, consumption_urls, \
    package_urls


class Dataset(SirenComponent):

    def get_s3_access_keys_for_dataset(self, *dataset_ids):
        """
        Retrieve S3 access keys for the specified account to access the
        specified dataset(s).

        :param list dataset_ids: One ore more dataset ids to get access to.
        :returns: A namedtuple containing the AWS keys and session token.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                s3_access_keys = client.get_s3_access_keys_for_dataset(dataset_id1, dataset_id2)
                # print(s3_access_keys)
                # access_key(access_key_id='39D19A440AFE452B9', secret_access_key='F426A93CDECE45C9BFF8F4F19DA5CB81', session_token='C0CC405803F244CA99999')

        """
        payload = {"datasetIds": list(dataset_ids)}

        response = self.session.post(
            dataset_urls.access_keys, json=payload
        )

        return siren_to_entity(response.to_siren())

    def get_dataset_partitions(self, id=None, name=None, package_id=None, package_name=None):
        """
        Retrieves the list of available partitions for a given dataset

        :param str id: The id of the dataset.
        :param str name: The name of the dataset.
        :param str package_id: The id of the package to which this dataset belongs.
                            Either this or package name is required if dataset is being looked up by name.
        :param str package_name: The name of the package to which this dataset belongs.
                            Either this or package id is required if dataset is being looked up by name.

        :returns: The dataset partitions in JSON format ( e.g. { ‘location’: [‘us’, ‘eu’], ...} )
        :rtype: dict

        - **Sample**

        .. code-block:: python

                # Look up by dataset id
                dataset = client.get_dataset_partitions('my_dataset_id')
                # or equivalent
                dataset = client.get_dataset_partitions(id='my_dataset_id')

                # Look up by dataset name

                # If package id is known
                dataset = client.get_dataset_partitions(name='my_dataset', package_id='my_package_id')

                # if package name is known
                dataset = client.get_dataset_partitions(name='my_dataset', package_name='my_package')
        """
        if name and (package_name or package_id):
            warnings.warn(
                'Getting a dataset by name, and package name or package ID'
                'will be deprecated in future. Slugs will replace this.',
                PendingDeprecationWarning
            )

            id = self._get_dataset_id_by_name(
                dataset_name=name,
                package_id=package_id,
                package_name=package_name
            )

        ds = self.get_dataset(id=id)
        response = ds.partitions()

        return response

    def get_dataset(self, id=None, name=None, package_id=None, package_name=None,
            dataset_short_code=None, organisation_short_code=None):
        """
        Retrieves a dataset.

        :param str id: The id of the dataset.
        :param str name: The name of the dataset.
        :param str package_id: The id of the package to which this dataset belongs.
                            Either this or package name is required if dataset is being looked up by name.
        :param str package_name: The name of the package to which this dataset belongs.
                            Either this or package id is required if dataset is being looked up by name.
        :param str dataset_short_code: Optional field. The short code of the requested dataset.
                            Should be used with organisation_short_code.
        :param str organisation_short_code: Optional field.
                            The short code of the organisation the requested dataset belongs to.
                            Needs to be used together with dataset_short_code.
                            If not specified, and dataset_short_code is defined,
                            then datasets will be searched within requesting user organisation.

        :returns: The dataset.
        :rtype: Dataset

        - **Sample**

        .. code-block:: python

                # Look up by dataset short code
                dataset = client.get_dataset(dataset_short_code='my_dataset_id')
                # or
                dataset = client.get_dataset(dataset_short_code='my_dataset_short_code', organisation_short_code='some_org_short_code')
                # Look up by dataset id
                dataset = client.get_dataset('my_dataset_id')
                # or equivalent
                dataset = client.get_dataset(id='my_dataset_id')

                # Look up by dataset name

                # If package id is known
                dataset = client.get_dataset(name='my_dataset', package_id='my_package_id')

                # if package name is known
                dataset = client.get_dataset(name='my_dataset', package_name='my_package')
        """
        if dataset_short_code:
            response = self.session.get(
                dataset_urls.v2_by_short_code.format(
                    dataset_short_code=dataset_short_code
                ), params={'organisation_short_code': organisation_short_code}
            )
        else:
            warnings.warn(
                'Getting a dataset by id or name with package name or package ID '
                'will be deprecated in future. Short codes should be used instead - '
                'please use dataset_short_code and organisation_short_code '
                'key arguments.',
                PendingDeprecationWarning
            )
            if not id and name and (package_name or package_id):
                id = self._get_dataset_id_by_name(
                    dataset_name=name,
                    package_id=package_id,
                    package_name=package_name
                )

            response = self.session.get(
                dataset_urls.v2_instance.format(id=id)
            )

        return self.Dataset._from_v2_response(response.json())

    def _get_dataset_id_by_name(self, package_id=None,
                                dataset_name=None,
                                package_name=None):

        if dataset_name and package_id:
            params = {'filter[0][field]': 'name',
                      'filter[0][operator]': 'eq',
                      'filter[0][value]': dataset_name}

            response = self.session.get(package_urls.package_datasets.format(
                id=package_id
            ), params=params)
            return response.json()['entities'][0]['properties']['datasetId']

        elif dataset_name and package_name:
            params = {"name": package_name}

            package_response = self.session.get(
                package_urls.package_index,
                params=params)

            package_id = package_response.json()["properties"]["packageId"]

            params = {
                'filter[0][field]': 'name',
                'filter[0][operator]': 'eq',
                'filter[0][value]': dataset_name
            }

            response = self.session.get(
                package_urls.v2_package_datasets.format(id=package_id),
                params=params
            )
            return response.json()["data"][0]["id"]

        else:
            raise Exception("No such package")

    def register_dataset(self, builder: DatasetBuilder):
        """
        Submit a request to create a new dataset under a specified package in the Data Catalogue.

        :param dli.client.builders.DatasetBuilder builder: An instance of DatasetBuilder. This builder object sets sensible defaults and exposes
                                                           helper methods on how to configure its storage options.

        :returns: A newly created Dataset.
        :rtype: Dataset

        - **Sample**

        .. code-block:: python

                # Please refer to builder docs for examples on
                # how to create an instance of DatasetBuilder.

                dataset = client.register_dataset(builder)
        """
        mp_encoder = builder.to_multipart_body()
        result = self.session.post(
            dataset_urls.v2_index,
            data=mp_encoder,
            headers={
                'Content-Type': mp_encoder.content_type,
                'Content-Length': str(mp_encoder.len),
            },
        ).json()

        return self.Dataset._from_v2_response(result)

    def edit_dataset(
        self,
        dataset_id,
        location_builder=None,
        **kwargs
    ):
        """
        Updates information on a dataset, returning the updated instance.
        If keyword argument is not specified field keeps its old value.
        Optional enum and text type fields can be unset by passing ``None``.

        :param str dataset_id: Id of the dataset being updated.
        :param dli.client.builders.DatasetLocationBuilder location_builder: Optional. An instance of DatasetLocationBuilder. This builder object exposes
                                                                            helper methods to configure dataset storage options.

        Keyword arguments:

        :keyword str name: Optional. A descriptive name of a dataset. It should be unique within the package.
        :keyword str description: Optional. A short description of a package.
        :keyword str content_type: Optional. A way for the data steward to classify the type of data in the dataset
                                (e.g. pricing).
        :keyword str data_format: Optional. The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `TSV`, XML`, `Other`.
        :keyword str publishing_frequency: Optional. The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
        :keyword list[str] keywords: Optional. User-defined list of keywords that can be used to find this dataset through
                                    the search interface.
        :keyword str naming_convention: Optional. Key for how to read the dataset name.
        :keyword str documentation: Optional. Documentation about this dataset in markdown format.
        :keyword list[str] taxonomy: Optional. A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
        :keyword str load_type: Optional. Whether each datafile in this dataset should be considered as a full version
                                of a dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`
        :keyword str data_preview_type: Optional enum of NONE STATIC LIVE. 'NONE' means there is no datapreview. STATIC means
                            to use the file provided. LIVE means to use the latest datafile.
        :keyword file sample_data: Optional file. File like object. Used in conjuntion with data_preview_type. Max size 10Megabytes.

        :returns: Updated Dataset.
        :rtype: Dataset

        - **Sample**

        .. code-block:: python

                # e.g. update dataset description
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    description="Updated my dataset description"
                )

                # update dataset location. Please note that this is only allowed if the dataset has no datafiles registered.
                builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    location_builder=builder
                )

                # update dataset taxonomy
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    taxonomy=["Credit", "CDS"]
                )

        """
        builder = DatasetBuilder(
            location_builder=location_builder,
            **kwargs
        )

        mp_encoder = builder.to_multipart_body()
        result = self.session.patch(
            dataset_urls.v2_instance.format(id=dataset_id), data=mp_encoder,
            headers={
                'Content-Type': mp_encoder.content_type,
                'Content-Length': str(mp_encoder.len),
            }
        ).json()

        return self.Dataset._from_v2_response(result)

    def delete_dataset(self, dataset_id):
        """
        Marks a particular dataset (and all its datafiles) as deleted.
        This dataset will no longer be accessible by consumers.

        :param str dataset_id: The id of the dataset to be deleted.

        :returns: None

        - **Sample**

        .. code-block:: python

                client.delete_dataset(dataset_id)

        """
        self.session.delete(
            dataset_urls.v2_instance.format(id=dataset_id)
        )

    def get_datafiles(self, dataset_id, name_contains=None,
                      as_of_date_start=None, as_of_date_end=None,
                      page=1, count=100):
        """
        Returns a list of all datafiles registered under a dataset.

        :param str dataset_id: The id of the dataset.
        :param str name_contains: Optional. Look up only those datafiles for the dataset where name contains this string.
        :param str as_of_date_start: Optional. Datafiles having data_as_of date greater than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param str as_of_date_end: Optional. Datafiles having data_as_of date less than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param int count: Optional count of datafiles to be returned. Defaults to 100.

        :returns: List of all datafiles registered under the dataset.
        :rtype: List[Datafile]

        - **Sample**

        .. code-block:: python

                datafiles = client.get_datafiles(
                    dataset_id,
                    name_contains='My Test Data',
                    as_of_date_start='2018-10-11',
                    as_of_date_end='2018-10-15',
                    count=10
                )
        """
        ensure_count_is_valid(count)

        params = {
            'name': name_contains,
            'as_of_date_start': as_of_date_start,
            'as_of_date_end': as_of_date_end,
            'page_size': count,
            'page': page,
        }

        return self.session.get(
            dataset_urls.datafiles.format(id=dataset_id), params=params
        ).to_many_siren('datafile')

    def get_latest_datafile(self, dataset_id):
        """
        Fetches datafile metadata of latest datafile in the dataset.

        :param str dataset_id: The id of the dataset.

        :returns: The datafile.
        :rtype: Datafile

        - **Sample**

        .. code-block:: python

                datafile = client.get_latest_datafile(dataset_id)
        """
        datafile = self.session.get(
            dataset_urls.latest_datafile.format(id=dataset_id)
        ).to_siren()

        return siren_to_entity(datafile)


class Dictionary(SirenComponent):

    def _register_schema(self, dataset_id, payload):
        response = self.session.post(
            dataset_urls.dictionary_index, json={'data': {'attributes': payload}}
        )

        return DictionaryModel(response.json()['data'], client=self)

    def register_dictionary(
        self,
        dataset_id,
        version,
        valid_as_of,
        fields,
        **kwargs
    ):
        """
        Registers dictionary metadata for a dataset.

        :param str dataset_id: Id of the dataset for the dictionary.
        :param str version: A user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: The date as of which the dictionary is active.
                               Expected format is YYYY-MM-DD. Must be unique.
        :param list[dict] fields: Non empty list of `Field` as described below.
        :param list[dict] partitions: Optional. Non empty list of `Partition` as described below.
        :param str description: Optional. Description for the dictionary.

        :returns: The registered dictionary

        Types
        =====

        Dictionary Field:

        .. code-block:: python

            {
                name	        string  - required
                type	        string  - required
                nullable        boolean - required
                metadata	dictionary
                description	string
                sample_value	string
                short_name	string
                is_derived	boolean
                validation	string
                comment	        string
            }

        Partition:

        .. code-block:: python

            {
                name: string,
                type: string
            }

        - **Sample**

        .. code-block:: python

                my_dictionary_fields = [
                            {
                                'name': 'field_1',
                                'type': 'String',
                                'nullable': False
                            },
                            {
                                'name': 'field_2',
                                'type': 'Double',
                                'nullable': False
                            },
                            {
                                'name': 'field_3',
                                'type': 'Int',
                                'nullable': True,
                                'metadata': {
                                    'some_key': 'some_value'
                                }
                            },
                        ]
                my_dictionary_partitions = [
                    {
                        'name': 'field_1',
                        'type': 'String'
                    }
                ]

                my_dictionary = client.register_dictionary(
                    "my-dataset-id",
                    version='1a',
                    valid_as_of='2018-10-31',
                    fields=my_dictionary_fields,
                    partitions=my_dictionary_partitions,
                    description="My dictionary description"
                )
        """
        payload = {
            'dataset_id': dataset_id,
            'version': version,
            'valid_as_of': valid_as_of,
            'fields': fields,
        }

        payload.update(**kwargs)

        response = self.session.post(
            dataset_urls.dictionary_index, json={'data': {'attributes': payload}}
        )

        return DictionaryModel(response.json()['data'], client=self)

    def register_spark_schema(
        self,
        dataset_id,
        version,
        valid_as_of,
        spark_schema_json,
        partitions=None,
        description=None,
        **kwargs
    ):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.register_schema` instead.

            Please migrate this function to register_schema. There is no longer an
            difference in functionality.

            Please migrate this:

                client.register_spark_schema(
                    '1234',
                    version='1a',
                    valid_as_of='2000-01-01',
                    spark_schema_json={}
                )

            To:

                client.register_schema(
                    '1234',
                    version='1a',
                    valid_as_of='2000-01-01',
                    **spark_schema_json
                )

        Registers dataframe schema for the datalake `dataset` for instances where data ingest pipelines make use of spark.

        :param str dataset_id: Id of the dataset for the schema.
        :param str version: A user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: The date as of which the schema is active.
                               Expected format is YYYY-MM-DD.
        :param dict spark_schema_json: Json schema for the spark dataframe.
        :param list[dict] partitions: Optional. Non empty list of `Partition` as described below.
        :param str description: Optional. Description for the schema.

        .. code-block:: python

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }

        :returns: The registered schema.

        - **Sample**

        .. code-block:: python

                # my_spark_dataframe: some spark dataframe with data being published in the data lake

                my_spark_schema_json = my_spark_dataframe.schema.jsonValue()

                my_schema = client.register_spark_schema(
                    "my-dataset-id",
                    version='1a',
                    valid_as_of='2018-10-31',
                    spark_schema_json=my_spark_schema_json,
                    description="My schema description"
                )
        """
        warnings.warn(
            '''
            `register_spark_schema` is deprecated. Please migrate this
            function to register_schema. There is no longer an
            difference in functionality. Please refer the deprecation documentation.

            https://docs.datalake.ihsmarkit.com/api_docs/index.html#dli.client.components.dataset.Dictionary.register_spark_schema
            ''',
            DeprecationWarning
        )

        payload = {
            'dataset_id': dataset_id,
            'version': version,
            'valid_as_of': valid_as_of,
        }

        payload.update(spark_schema_json)
        payload.update(kwargs)

        return self.register_dictionary(**payload)

    def get_dictionary(self, dataset_id, version=None):
        """
        Looks up dictionary for a dataset by version. In case version is not specified this will fetch dictionary version having the latest valid_as_of date.
        Throws exception if no dictionary is registered for the dataset.

        :param str dataset_id: The id of the dataset under which the dictionary is registered.
        :param str version: Optional. The version of the dictionary.

        :returns: The dictionary.

        - **Sample**

        .. code-block:: python

                # Fetch dictionary version '1a'
                dictionary = client.get_dictionary('my_dataset_id', version='1a')

                # Fetch the dictionary with the latest valid_as_of date
                latest_dictionary = client.get_dictionary('my_dataset_id')

        """

        if version:
            schema = self.session.get(
                dataset_urls.v2_schema_instance_version.format(
                    id=dataset_id,
                    version=version
                )
            )
        else:
            schema = self.session.get(
                dataset_urls.dictionary_by_dataset_lastest.format(
                    id=dataset_id
                )
            )
        json = schema.json()['data']
        return DictionaryModel(json, client=self)

    def get_dictionaries(self, dataset_id, count=100):
        """
        Returns a list of all dictionaries registered under the dataset. The list is sorted by dictionary valid_as_of date in descending order.

        :param str dataset_id: The id of the dataset.
        :param int count: Optional count of dictionaries to be returned. Defaults to 100.

        :returns: List of all dictionaries registered under the dataset sorted by valid_as_of date in descending order.

        - **Sample**

        .. code-block:: python

                my_dictionaries = client.get_dictionaries("my_dataset_id", count=10)
        """
        response = self.session.get(
            dataset_urls.dictionary_by_dataset.format(id=dataset_id)
        )

        return [
            DictionaryModel(d, client=self) for d in
            response.json()['data']
        ][:count]  # todo - will kill this silly count stuff ASAP

    def delete_dictionary(self, dataset_id, version):
        """

        Marks a dictionary version for a dataset as deleted.

        :param str dataset_id: The id of the dataset under which the dictionary is registered.
        :param str version: The version of the dictionary.

        :returns: None

        - **Sample**

        .. code-block:: python

                # Delete dictionary version '1a'
                client.delete_dictionary(dataset_id='my_dataset_id', version='1a')

        """
        schema = self.get_dictionary(dataset_id, version)
        self.session.delete(
            dataset_urls.dictionary_instance.format(id=schema.id)
        )

    def edit_dictionary(
        self,
        dataset_id,
        version,
        new_version=None,
        **kwargs
    ):
        """
        Updates dictionary metadata for a dataset.
        If a field is passed as ``None`` then the field will not be updated.

        :param str dataset_id: Id of the dataset for the dictionary_instance.
        :param str version: Version of the dictionary being updated.
        :param str new_version: Optional. New version if to be updated. This is a user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: Optional. The date as of which the dictionary is active.
                               Expected format is YYYY-MM-DD.
        :param list[dict] fields: Optional. If provided, a non empty list of `Field` as described below.
        :param list[dict] partitions: Optional. If provided, a non empty list of `Partition` as described below.
        :param str description: Optional. Description for the dictionary.

        .. code-block:: python

                # Field:
                {
                    "name": "field_a", 			# name of the column.
                    "nullable": True,  			# defaulted to True - A boolean indicating whether the field is nullable or not.
                    "metadata": None			# optional dictionary with metadata for this column.
                }

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }

        :returns: The updated dictionary.
        :rtype: dli.models.dictionary_model.DictionaryModel

        - **Sample**

        .. code-block:: python

                # Updating description and valid_as_of date for my dictionary
                my_updated_schema = client.edit_dictionary(
                    "my-dataset-id",
                    '1a',
                    valid_as_of='2018-11-05',
                    description="My updated dictionary description"
                )
        """
        schema = self.get_dictionary(dataset_id, version)
        payload = dict(**kwargs)

        if new_version is not None:
            payload['version'] = new_version

        response = self.session.patch(
            dataset_urls.dictionary_instance.format(id=schema.id),
            json={'data': {'attributes': payload}}
        )

        return DictionaryModel(response.json()['data'], client=self)

    def delete_schema(self, *args, **kwargs):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.delete_dictionary` instead.
        """
        warnings.warn(
            '.delete_schema is deprecated. Please use .delete_dictionary',
            DeprecationWarning
        )

        return self.delete_dictionary(*args, **kwargs)

    def edit_schema(self, *args, **kwargs):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.edit_dictionary` instead.
        """
        warnings.warn(
            '.edit_schema is deprecated. Please use .edit_dictionary',
            DeprecationWarning
        )

        return self.edit_dictionary(*args, **kwargs)

    def get_schema(self, *args, **kwargs):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.get_dictionary` instead.
        """
        warnings.warn(
            '.get_schema is deprecated. Please use .get_dictionary',
            DeprecationWarning
        )
        return self.get_dictionary(*args, **kwargs)

    def get_schemas(self, *args, **kwargs):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.get_dictionaries` instead.
        """
        warnings.warn(
            '.get_schemas is deprecated. Please use .get_dictionaries',
            DeprecationWarning
        )

        return self.get_dictionaries(*args, **kwargs)

    def register_schema(self, *args, **kwargs):
        """
        .. deprecated:: 1.1.0
           Use :func:`dli.client.components.dataset.Dictionary.register_dictionary` instead.
        """
        warnings.warn(
            '.register_schema is deprecated. Please use .register_dictionary',
            DeprecationWarning
        )

        return self.register_dictionary(*args, **kwargs)
